#%%
import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np
from scipy.stats import norm

mpl.rcParams["figure.dpi"] = 300

from sklearn.mixture import GaussianMixture


#%%
def h_index(list_distributions) -> float:
    """
    Computes the h-index given a mixture of ditributions.

    Parameters
    ----------
    list_distributions : list
        List of np.arrays containing estimated subdistributions

    Returns
    -------
    float
        calculated H-index
        
    Notes
    -----    
    When comparing H-index between datasets be sure to set the n_components
    paramter of the GaussianMixtureModel to the same value.
       
    References
    ----------
    http://www.microscopist.co.uk/wp-content/uploads/2021/06/FLIM-review.pdf

    """
    median_overall = np.median(np.concatenate(list_distributions))

    # compute running sum
    running_sum = 0
    for distribution in list_distributions:
        pass
        p = len(distribution) / np.sum(
            list(map(len, list_distributions))
        )  # proportion of subpopulation
        d = np.abs(
            np.mean(distribution) - median_overall
        )  # distance between median of subpop and overall med
        running_sum += d * p * np.log(p)
    return -running_sum


#%%


def h_index_single_weighted(list_distributions) -> float:
    """
    Computes the single weighted h-index given a mixture of ditributions.

    Parameters
    ----------
    list_distributions : list
        List of np.arrays containing estimated subdistributions

    Returns
    -------
    float
        calculated H-index
        
    Notes
    -----
    When comparing H-index between datasets be sure to set the n_components
    paramter of the GaussianMixtureModel to the same value.

    """
    median_overall = np.mean(np.concatenate(list_distributions))

    # compute running sum
    running_sum = 0
    for distribution in list_distributions:
        pass
        p = len(distribution) / np.sum(
            list(map(len, list_distributions))
        )  # proportion of subpopulation
        d = np.abs(
            np.mean(distribution) - median_overall
        )  # distance between median of subpop and overall med
        sigma = np.std(distribution)
        rng = np.random.default_rng(seed=1)
        im = rng.random((128, 128))
        running_sum += (1 - (p * np.log(p + 1))) * (sigma + d)

    return running_sum


#%%
if __name__ == "__main__":
    pass
    #%% GENERATE DISTRIBUTIONS FOR TESTING
    x = np.linspace(1, 100, num=1000)

    # distributions to create
    list_dist_params = [
        # (mean, std)
        (20, 4),
        # (20,6),
        # (20,10),
        # (20,10),
        (20, 4),
        # (30,6),
        (50, 20),
        (50, 10),
    ]

    # list_dist_params = [
    #    #(mean, std)

    #     ]

    dist_all = np.zeros_like(x)
    list_distributions = []

    for dist_params in list_dist_params:
        pass
        mean, sigma = dist_params
        dist = norm(loc=mean, scale=sigma)
        list_distributions.append(dist)
        dist_all = dist_all + dist.pdf(x)

    # plot dataset
    plt.title("Distributions")
    for pos, dist in enumerate(list_distributions, start=1):
        pass
        plt.plot(
            x,
            dist.pdf(x),
            label=f"dist {pos} | mean: {dist.mean()} | stdev: {dist.std()}",
        )
    plt.plot(x, dist_all, label="all distributions")
    plt.xlabel("value")
    plt.ylabel("probability")
    plt.legend()
    plt.show()

    ## try out function
    # generate random values
    list_dist_values = [d.rvs(len(x)) for d in list_distributions]

    list_dist_values[0].mean()
    plt.tight_layout()
    # h_index_value = h_index(list_distributions)
    # print(h_index_value)

    #%% COMBINE DISTRIBUTIONS AND COMPUTE WEIGHTED AND SINGLY WEIGHTED H-INDEX

    ## plot one two three distributions
    n_components = 3

    plt.title(f"Distributions \nGaussian Mixture Model components : {n_components} ")

    for num_dist in np.arange(start=1, stop=5):  # slide in dist
        pass
        list_subdists = list_distributions[:num_dist]

        # compute h_index
        list_subdists_values = [dist.rvs(len(x)) for dist in list_subdists]

        # compute gaussiam mixture model
        all_data = np.array(list_subdists_values).reshape(-1, 1)
        # plt.hist(all_data, histtype="step")
        # plt.show()

        # fit data into gaussian mixtures
        # https://towardsdatascience.com/gaussian-mixture-model-clusterization-how-to-select-the-number-of-components-clusters-553bef45f6e4
        # https://stats.stackexchange.com/questions/139163/selecting-the-number-of-mixtures-hidden-states-latent-variables
        # https://en.wikipedia.org/wiki/Determining_the_number_of_clusters_in_a_data_set
        gauss = GaussianMixture(n_components=n_components)
        labels = gauss.fit_predict(all_data)

        list_fit_dist = [
            np.asarray(all_data[labels == label]).flatten()
            for label in np.unique(labels)
        ]

        ## GAUSSIAN MIXTURE MODEL
        # plt.title(f"histogram of fit data | fixed GMM | components: {len(np.unique(labels))}")
        # for list_data in list_fit_dist:
        #     plt.hist(list_data, bins=100, histtype='step', label=f'mean: {np.mean(list_data):.3f} | weight:{(len(list_data)/len(all_data)):.2f}')
        #     plt.ylabel("count")
        #     plt.xlabel("value")
        # plt.legend()
        # plt.show()

        ##### BAYESIAN GAUSSIAN MIXTURE MODEL

        # from sklearn.mixture import BayesianGaussianMixture
        # bgmm = BayesianGaussianMixture(n_components=4,
        #                                max_iter=1000
        #                             )
        # model_labels = bgmm.fit_predict(all_data)
        # list_fit_dist = [np.asarray(all_data[model_labels == label]).flatten() for label in np.unique(model_labels)]

        # plt.title(f"histogram of fit data Bayesian | components: {len(np.unique(model_labels))} ")
        # for idx, list_data in enumerate(list_fit_dist):
        #     pass
        #     plt.hist(list_data, bins=100, histtype='step', label=f'{idx} | {np.mean(list_data):.3f} | weight:{(len(list_data)/len(all_data)):.2f}')
        #     plt.ylabel("count")
        #     plt.xlabel("value")
        # plt.legend()
        # plt.show()

        #############

        h_index_value = h_index(list_fit_dist)
        h_index_single_weighted_value = h_index_single_weighted(list_fit_dist)

        # plot pdf
        pdf_combined = [d.pdf(x) for d in list_subdists]
        pdf_combined_values = np.sum(pdf_combined, axis=0)
        plt.plot(
            x,
            pdf_combined_values,
            label=f"dist {num_dist} | Hi: {h_index_value:.2f} | wHi: {h_index_single_weighted_value:.2f}",
        )
    plt.legend()
    plt.show()
# %%



import numpy as np


def _validate_array_and_make_bool(mask):
    """
    Function to validate that the mask used is boolean and that it has data
    """

    # convert to bool and remove 1D axis
    mask = mask.astype(bool).squeeze()

    # check that it's 2d
    n_dims = len(mask.shape)

    if n_dims != 2:
        raise ValueError(f"Shape Error: mask should be 2D, {n_dims} dimensions found")

    if np.sum(mask) == 0:
        # check that mask has data
        raise ValueError("Error: no data found in mask")

    return mask



# Dependencies
import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def lifetime_to_phasor(f, timebins, counts):
    """ Time to frequency domain transformation

    Parameters
    ---------- 
        f  : int 
            laser repetition angular frequency
        timebins : ndarray
            numpy array of timebins
        counts : ndarray
            photon counts of the histogram()

    Returns
    -------
        angle  : float 
            angle in radians
        magnitude  : float 
            magnitude of phasor
    """
    w = 2 * np.pi * f  # f
    phasor = coll.namedtuple("phasor", "angle magnitude")  # nameddtuple
    # pylab.plot(timebins,counts)

    ## convert to phasor rectangular
    point_g = np.sum(counts * np.cos(w * timebins)) / np.sum(counts)
    point_s = np.sum(counts * np.sin(w * timebins)) / np.sum(counts)

    # https://software.intel.com/en-us/forums/archived-visual-fortran-read-only/topic/313067
    # 0.5*TWOPI-ATAN2(Y,-X)
    # angle = AMOD(ATAN2(y,x)+TWOPI,TWOPI)
    angle = np.pi - np.arctan2(point_s, -point_g)
    magnitude = np.sqrt(point_g ** 2 + point_s ** 2)

    return phasor(angle=angle, magnitude=magnitude)


if __name__ == "__main__":
    pass



from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt
import tifffile

mpl.rcParams["figure.dpi"] = 300
import re

import numpy as np
from scipy.ndimage import grey_dilation, grey_erosion
from skimage import morphology
from sklearn.model_selection import ParameterGrid

from cell_analysis_tools.image_processing import bin_im, kmeans_threshold, normalize
from cell_analysis_tools.metrics import dice

if __name__ == "__main__":

    # load image
    path_images = Path(
        r"Z:\0-Projects and Experiments\TQ - AAG - mitochondria-actin segmentation\210205_GoldStandard_NADH_images".replace(
            "\\", "/"
        )
    )
    list_path_images = list(path_images.glob("*.tif"))
    path_image = list_path_images[0]
    im = tifffile.imread(path_image)
    # plt.imshow(im)
    # plt.show()

    # load mask
    path_masks = Path(
        r"Z:\0-Projects and Experiments\TQ - AAG - mitochondria-actin segmentation\210111_GoldStandard_Stain_Masks".replace(
            "\\", "/"
        )
    )
    path_masks_aag = path_masks / "AAG-masks"
    path_masks_tq = path_masks / "TQ-masks"
    list_path_masks_aag = list(path_masks_aag.glob("*.tif"))
    list_path_masks_tq = list(path_masks_tq.glob("*.tif"))

    list_path_masks_aag_str = [str(p) for p in list_path_masks_aag]
    list_path_masks_tq_str = [str(p) for p in list_path_masks_tq]

    # string to find
    handle_nadh = path_image.stem

    if handle_nadh.split("_", 1)[0] == "2020":
        handle_tmre_prefix = handle_nadh.rsplit("_", 3)[0]
        handle_tmre_suffix = handle_nadh.split("_", 5)[-1]
        handle_tmre = handle_tmre_prefix + "_TMRE_" + handle_tmre_suffix

    # mask
    path_mask_aag = list(
        filter(re.compile(f"{handle_tmre}").search, list_path_masks_aag_str)
    )[0]
    path_mask_tq = list(
        filter(re.compile(f"{handle_tmre}").search, list_path_masks_tq_str)
    )[0]

    mask_aag = tifffile.imread(path_mask_aag)
    mask_tq = tifffile.imread(path_mask_tq)

    # normalize
    im = normalize(im)

    list_bin_sizes = np.arange(0, 10, step=1)
    list_disk_radius = np.arange(0, 10, step=1)

    param_grid = {
        "bin_size": list_bin_sizes,
        "selem_radius": list_disk_radius,
        # "kmeans_"
    }
    parameters = list(ParameterGrid(param_grid))

    # dice score placeholder --> bin_sizes in rows, disk radius in cols
    array_dice_scores_aag = np.zeros((len(list_bin_sizes), len(list_disk_radius)))
    array_dice_scores_tq = np.zeros((len(list_bin_sizes), len(list_disk_radius)))

    #%%

    for pos, params in enumerate(parameters, start=1):
        print(f"iteration {pos}/{len(parameters)}")
        bin_size = params["bin_size"]
        selem_radius = params["selem_radius"]

        # tophat with bin size zero doesn't make sense
        # if bin_size == 0:
        #     continue

        # bin_size=2
        kernel_size = bin_size * 2 + 1
        im_binned = bin_im(im, bin_size)
        # plt.imshow(im_binned)
        # plt.show()

        # tophat for bg equalization and edge sharpening
        # selem_radius = 11
        selem = morphology.disk(selem_radius)
        top_hat = morphology.white_tophat(im_binned, selem)
        # plt.imshow(top_hat)
        # plt.show()

        # manual top hat
        im_erosion = grey_erosion(im_binned, footprint=selem)
        # plt.imshow(im_erosion)
        # plt.show()

        im_dilation = grey_dilation(im_erosion, footprint=selem)
        # plt.imshow(im_dilation)
        # plt.show()

        im_subtraction = im_binned - im_dilation
        # plt.imshow(im_subtraction)
        # plt.show()

        assert not (
            top_hat - im_subtraction
        ).any(), "top_hat does not match manual top hat"

        if im_subtraction.sum() == 0:
            print(
                f"top_hat returned all zeros: bin_size:{bin_size} radius: {selem_radius}"
            )
            continue

        # cluster
        k = 2
        n_brightest_clusters = 1
        im_kmeans = kmeans_threshold(
            top_hat, k=k, n_brightest_clusters=n_brightest_clusters
        )

        # plt.title(f"bin: {kernel_size}x{kernel_size}  disk_radii:{selem_radius}")
        # plt.imshow(im_kmeans)
        # plt.show()

        debug = False
        if debug:
            # original    aag_mask   tq_mask
            # mask_pred   dice plot
            fig, ax = plt.subplots(2, 3, figsize=(20, 10))

            fig.suptitle(
                f"bin: {kernel_size}x{kernel_size}  disk_radius:{selem_radius}"
            )
            ax[0, 0].set_title(f"original")
            ax[0, 0].imshow(im)
            ax[0, 0].set_axis_off()

            ax[0, 1].set_title(f"mask_aag")
            ax[0, 1].imshow(mask_aag)
            ax[0, 1].set_axis_off()

            ax[0, 2].set_title(f"mask_tq")
            ax[0, 2].imshow(mask_tq)
            ax[0, 2].set_axis_off()

            ax[1, 0].set_title(f"gray_erosion")
            ax[1, 0].imshow(im_erosion)
            ax[1, 0].set_axis_off()

            ax[1, 1].set_title(f"gray_dilation")
            ax[1, 1].imshow(im_dilation)
            ax[1, 1].set_axis_off()

            ax[1, 2].set_title(f"top_hat \n (im - opening(im))")
            ax[1, 2].imshow(im_subtraction)
            ax[1, 2].set_axis_off()
            plt.show()

        array_dice_scores_aag[bin_size, selem_radius] = dice(im_subtraction, mask_aag)
        array_dice_scores_tq[bin_size, selem_radius] = dice(im_subtraction, mask_tq)

    #%%



import pathlib
from pathlib import Path
from typing import Union

import numpy as np
import tifffile

from .read_asc import read_asc


def load_image(path: Union[str, pathlib.PurePath]) -> np.ndarray:
    """
    Detects the extension and loads image into a numpy array 
    if it's a tif/tiff or an asc file.

    Parameters
    ----------
    path : pathlib path or str
        path to the image.

    Returns
    -------
    np.ndarray
        ndarray with the image data.

    """
    if not isinstance(path, pathlib.PurePath):
        path = Path(path)
    pass
    if path.suffix == ".asc":
        return read_asc(path)
    if path.suffix in [".tiff", ".tif"]:
        return tifffile.imread(path)



import numpy as np

from .helper import _validate_array_and_make_bool


def average_relative_performance(mask_gt1, mask_gt2, mask_pred):
    """
    Returns average relative performance of a predicted mask against 
    two ground truth masks from two different users
    
    Taken from mito hacker paper `https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7642274/`

    ((Mito Catcher vs. User 1 + Mito Catcher vs. User 2)/(User 1 vs. User 2))
    
    Parameters
    ----------
    mask_gt1 : ndarray
        ground truth mask 1.
    mask_gt2 : ndarray
        ground truth mask 2.
    mask_predicted : ndarray
        predicted mask.

    Returns
    -------
    float
        average performance score.

    """
    mask_gt1 = _validate_array_and_make_bool(mask_gt1)
    mask_gt2 = _validate_array_and_make_bool(mask_gt2)
    mask_predicted = _validate_array_and_make_bool(mask_pred)

    return (
        np.sum(np.logical_and(mask_predicted, mask_gt1))
        + np.sum(np.logical_and(mask_predicted, mask_gt2))
    ) / (2 * np.sum((np.logical_and(mask_gt1, mask_gt2))))



from .bin_image import bin_image
from .draw_universal_semicircle import draw_universal_semicircle
from .estimate_and_shift_irf import estimate_and_shift_irf
from .ideal_sample_phasor import ideal_sample_phasor
from .phasor_calibration import phasor_calibration
from .phasor_to_rectangular import phasor_to_rectangular
from .regionprops_omi import regionprops_omi
from .lifetime_to_phasor import lifetime_to_phasor
from .rectangular_to_phasor import rectangular_to_phasor
from .phasor_calculator import phasor_calculator

__all__ = [
    "bin_image",
    "draw_universal_semicircle",
    "estimate_and_shift_irf",
    "ideal_sample_phasor",
    "phasor_calibration",
    "regionprops_omi",
    "lifetime_to_phasor",
    "phasor_to_rectangular",
    "rectangular_to_phasor",
    'phasor_calculator'
    
]



from sklearn.metrics import roc_curve, auc,  accuracy_score, classification_report # confusion_matrix
from sklearn.ensemble import RandomForestClassifier

from sklearn.preprocessing import label_binarize
from sklearn.model_selection import train_test_split

import pandas as pd
from copy import deepcopy
import numpy as np

import matplotlib.pylab as plt
import matplotlib as mpl
mpl.rcParams['figure.dpi'] = 300

import seaborn as sns



#%%  ######## CLASSIFIER FUNCTION
def run_analysis_on_classifier(clf, X_test, y_test, dict_classes):
    print(clf)
    y_proba = clf.predict_proba(X_test)
    # y_pred  = clf.predict(X_test) # no control over operating point
    
    # check probab and y_pred 
    
    # bool_proba_pred_agreement = (y_pred == (y_proba[:,1] > 0.5)).all()
    # if bool_proba_pred_agreement:
    #     print("proba same as pred then thresholded at 0.5")
    # else:
    #     print(f"{'*' * 40} proba NOT the same as pred then thresholded at 0.5")
    
    actual_neg = y_test == 0
    actual_pos = y_test == 1
    print('=' * 41)
    print(f"{'Actual':^15} | Total : {len(y_test)} | CD68- : {sum(actual_neg)}  | CD68+ : {sum(actual_pos)}")
    prob_cd69pos = y_proba[:,1] # these are CD69+ probabilities
    

    # Compute ROC curve and AUC for each class
    print(f"y_test : {len(y_test)} | prob_cd69pos len : {len(prob_cd69pos)}")
    fpr, tpr, array_thresh = roc_curve(y_test, prob_cd69pos) # get label of CD69+, 2nd class
    roc_auc = auc(fpr, tpr)

    
    # ########## ECG trying out probability thresholds
    print('-' * 41, "Emmanuel analysis")
    threshold = 0.50
    pred_neg = prob_cd69pos <= threshold
    pred_pos = prob_cd69pos > threshold
        
    
    TPR = sum(pred_pos * actual_pos) / sum(actual_pos)
    FPR = sum(pred_pos * actual_neg) / sum(actual_neg)
    print(f"Threshold {threshold:.2f} |  CD69- : {sum(pred_neg)} | CD69+ {sum(pred_pos)} | TPR: {TPR:.3f} FPR: {FPR:.3f}")
    print('=' * 41)

    print('-' * 41, "python analysis")
    y_test = pd.Series(y_test.squeeze()).map(dict_classes)
    # y_pred = pd.Series(np.array(pred_pos).astype(int)).map(dict_classes)
    y_pred = pd.Series(np.array(pred_pos, dtype=np.int)).map(dict_classes)
    
    confusion_matrix = pd.crosstab(y_test,
                                   y_pred,
                                   rownames=['Actual'], 
                                   colnames=['Predicted']
                                   )
    print(confusion_matrix)
  
    #print metrics to assess classifier performance
    accuracy = accuracy_score(y_test, y_pred)
    print('Accuracy score =', accuracy)
    print(classification_report(y_test, y_pred))

    return fpr, tpr, roc_auc, accuracy, (FPR, TPR) # (operating_point_fpr, operating_point_tpr)#



# #%%  ######## CLASSIFIER FUNCTION
# def run_analysis_on_classifier(clf, X_test, y_test, dict_classes):
#     print(clf)
#     y_proba = clf.predict_proba(X_test)
#     # y_pred  = clf.predict(X_test) # no control over operating point
    
#     # check probab and y_pred 
    
#     # bool_proba_pred_agreement = (y_pred == (y_proba[:,1] > 0.5)).all()
#     # if bool_proba_pred_agreement:
#     #     print("proba same as pred then thresholded at 0.5")
#     # else:
#     #     print(f"{'*' * 40} proba NOT the same as pred then thresholded at 0.5")
    
#     actual_neg = y_test == 0
#     actual_pos = y_test == 1
#     print('=' * 41)
#     print(f"{'Actual':^15} | Total : {len(y_test)} | CD68- : {sum(actual_neg)}  | CD68+ : {sum(actual_pos)}")
#     prob_cd69pos = y_proba[:,1] # these are CD69+ probabilities
    

#     # Compute ROC curve and AUC for each class
#     print(f"y_test : {len(y_test)} | prob_cd69pos len : {len(prob_cd69pos)}")
#     fpr, tpr, array_thresh = roc_curve(y_test, prob_cd69pos) # get label of CD69+, 2nd class
#     roc_auc = auc(fpr, tpr)
#     # print(f"{'predicted':^15} |  CD69- : {sum(y_pred==0)} | CD69+ {sum(y_pred==1)} ")
#     # print(f" Total predicted : {len(y_pred)}  | fpr count: {len(fpr)} | tpr count: {len(tpr)}")
    
    
#     # ########## ECG trying out probability thresholds
#     print('-' * 41, "Emmanuel analysis")
#     threshold = 0.50
#     pred_neg = prob_cd69pos <= threshold
#     pred_pos = prob_cd69pos > threshold
    
#     # print((pred_pos == y_pred).all()) # check to see if predictions y sklearn same as >0.5 threshold
    
    
#     TPR = sum(pred_pos * actual_pos) / sum(actual_pos)
#     FPR = sum(pred_pos * actual_neg) / sum(actual_neg)
#     print(f"Threshold {threshold:.2f} |  CD69- : {sum(pred_neg)} | CD69+ {sum(pred_pos)} | TPR: {TPR:.3f} FPR: {FPR:.3f}")
#     print('=' * 41)

#     # ############## Kayvan - Cost function
#     # print('-' * 41, "Kayvan analysis")
#     # # idx = np.argwhere(array_thresh ==0.5)
#     # # cost = fpr - tpr # 45 degree line adjacent to roc, top left down
    
#     # # weighted by number of members of each class 
#     # # minimizes off diagonal of confusion matrix, fp, fn misclassifications
#     # cost = fpr * sum(actual_neg) - tpr * sum(actual_pos) 
#     # idx_min = np.argmin(cost)
    
#     # thresh_optimum = array_thresh[idx_min]
#     # operating_point_fpr = fpr[idx_min]
#     # operating_point_tpr = tpr[idx_min]
#     # # print(f"fpr: {operating_point_fpr:.3f} | tpr : {operating_point_tpr:.3f} | operating threshold: {thresh_optimum}")
    
#     # # scores_cd69 = y_proba[:,1] # these are CD69+ probabilities
#     # pred_neg = prob_cd69pos < thresh_optimum
#     # pred_pos = prob_cd69pos >= thresh_optimum
#     # print(f"Threshold {thresh_optimum:.2f} |  CD69- : {sum(pred_neg)} | CD69+ {sum(pred_pos)} | | TPR: {operating_point_tpr:.3f} FPR: {operating_point_fpr:.3f}")
#     # ##############

#     print('-' * 41, "python analysis")
#     y_test = pd.Series(y_test.squeeze()).map(dict_classes)
#     # y_pred = pd.Series(np.array(pred_pos).astype(int)).map(dict_classes)
#     y_pred = pd.Series(np.array(pred_pos, dtype=np.int)).map(dict_classes)
    
#     confusion_matrix = pd.crosstab(y_test,
#                                    y_pred,
#                                    rownames=['Actual'], 
#                                    colnames=['Predicted']
#                                    )
#     print(confusion_matrix)
  
#     # compute FPR and TPR
#     # tp = confusion_matrix.iloc[0,0]
#     # fp = confusion_matrix.iloc[1,0]
#     # fn = confusion_matrix.iloc[0,1]
#     # tn = confusion_matrix.iloc[1,1]
    
#     # TPR = tp / (tp + fn)
#     # FPR = fp/ (fp + tn)
    
#     # print(f"Threshold default |  CD69- : {tn + fn} | CD69+ {tp + fp} | TPR: {TPR:.3f} FPR: {FPR:.3f}")
    
#     # TODO fix this list
#     # for col, feature in zip(np.flip(df.columns[np.argsort(clf.feature_importances_)]), 
#     #                         np.flip(np.argsort(clf.feature_importances_))):
#     #     print(col, clf.feature_importances_[feature])

#     #print metrics to assess classifier performance
#     accuracy = accuracy_score(y_test, y_pred)
#     print('Accuracy score =', accuracy)
#     print(classification_report(y_test, y_pred))

#     return fpr, tpr, roc_auc, accuracy, (FPR, TPR) # (operating_point_fpr, operating_point_tpr)#



# https://python.plainenglish.io/solve-graph-coloring-problem-with-greedy-algorithm-and-python-6661ab4154bd

# Adjacent Matrix
G = [
    [0, 1, 1, 0, 1, 0],
    [1, 0, 1, 1, 0, 1],
    [1, 1, 0, 1, 1, 0],
    [0, 1, 1, 0, 0, 1],
    [1, 0, 1, 0, 0, 1],
    [0, 1, 0, 1, 1, 0],
]


# inisiate the name of node.
node = "abcdef"
t_ = {}
for i in range(len(G)):
    t_[node[i]] = i

# count degree of all node.
degree = []
for i in range(len(G)):
    degree.append(sum(G[i]))

# inisiate the posible color
colorDict = {}
for i in range(len(G)):
    colorDict[node[i]] = ["Blue", "Red", "Yellow", "Green"]


# sort the node depends on the degree
sortedNode = []
indeks = []

# use selection sort
for i in range(len(degree)):
    _max = 0
    j = 0
    for j in range(len(degree)):
        if j not in indeks:
            if degree[j] > _max:
                _max = degree[j]
                idx = j
    indeks.append(idx)
    sortedNode.append(node[idx])

# The main process
theSolution = {}
for n in sortedNode:
    setTheColor = colorDict[n]
    theSolution[n] = setTheColor[0]
    adjacentNode = G[t_[n]]
    for j in range(len(adjacentNode)):
        if adjacentNode[j] == 1 and (setTheColor[0] in colorDict[node[j]]):
            colorDict[node[j]].remove(setTheColor[0])


# Print the solution
for t, w in sorted(theSolution.items()):
    print("Node", t, " = ", w)



import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.flim import ideal_sample_phasor, lifetime_to_phasor
from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc

def phasor_calibration(f, lifetime, timebins, counts):
    """ Phasor Plot Calibration

    Parameters
    ---------- 
        f  : int 
            laser repetition angular frequency
        lifetime  : int 
            lifetime of known sample in ns (single exponential decay)
        timebins  : int 
            timebins of samples
        counts  : int 
            photon counts of histogram
    Returns
    -------
        angle_offset {float}: difference in angle between known sample and actual
        magnitude_offset {float}: difference in magnitude between known sample and actual
    Note
    ----
        If no timebins or histograms passed then returns angle and phase of
        decay passed in.
    """
    
    
    calibration = coll.namedtuple("calibration", "angle scaling_factor")

    # calculate idea and real phasors
    ideal_sampl_phasor = ideal_sample_phasor(f, lifetime)
    real = lifetime_to_phasor(f, timebins, counts)

    """ calculate angle offset """
    angle = ideal_sampl_phasor.angle - real.angle

    """ ratio of magnitudes -> ideal/actual """
    ratio = ideal_sampl_phasor.magnitude / real.magnitude

    return calibration(angle=angle, scaling_factor=ratio)


if __name__ == "__main__":
    
    from cell_analysis_tools.io import load_sdt_file
    from cell_analysis_tools.flim import (draw_universal_semicircle,
                                          phasor_to_rectangular,
                                          rectangular_to_phasor
                                          )
    
    
    # # load irf
    irf_decay = np.loadtxt("irf.csv")   
    timebins = irf_decay[:,0]
    irf = irf_decay[:,1]
    plt.plot(timebins,irf)
        
    # load and plot original irf
    irf_offset = np.roll(irf,100)
    f = 0.08 # Ghz
    phasor = lifetime_to_phasor(f=f, timebins=timebins, counts=irf)
    phasor = lifetime_to_phasor(f=f, timebins=timebins, counts=irf_offset)
    plt.plot(timebins, irf_offset)
    
    g,s = phasor_to_rectangular(phasor.angle, phasor.magnitude)
    draw_universal_semicircle(laser_angular_frequency=80e6)
    plt.scatter(g,s, label="original point")

    
    # calibate and plot original irf
    calibration = phasor_calibration(f=80e6,
                                    lifetime=0,
                                    timebins=timebins,
                                    counts=irf)    
    
    phasor_calibrated = phasor.angle + calibration.angle, phasor.magnitude * calibration.scaling_factor
    new_g, new_s = phasor_to_rectangular(phasor_calibrated[0], phasor_calibrated[1])
    
    # PLOT 
    draw_universal_semicircle(laser_angular_frequency=80e6,
                              suptitle=f"Transformation ofsset: angle = {calibration.angle:.2f} | scaling_factor = {calibration.scaling_factor:.2f}")
    plt.scatter(g,s, label="original point")
    plt.scatter(new_g, new_s, label="calibrated point")
    plt.legend()
    
    # plot bead data
    # load image 
    sdt = load_sdt_file("./bigger_beads_2.1ns.sdt")
    im_nadh = sdt.squeeze()
    decay = im_nadh.sum(axis=(0,1))
    
    phasor = lifetime_to_phasor(f=80e6, timebins=timebins, counts=decay)
    g,s = phasor_to_rectangular(phasor.angle, phasor.magnitude)
    plt.scatter(g, s)
    
    # calibrate 
    calibration = phasor_calibration(f=80e6,
                                            lifetime=2.2e-9,
                                            timebins=timebins,
                                            counts=decay)    
    
    phasor_calibrated = phasor.angle + calibration.angle, phasor.magnitude * calibration.scaling_factor
    new_g, new_s = phasor_to_rectangular(phasor_calibrated[0], phasor_calibrated[1])
    plt.scatter(new_g, new_s)
    
    phasor2 = rectangular_to_phasor(g, s)

    #### plot image
    plt.show()
    


import matplotlib as mpl
import numpy as np

mpl.rcParams["figure.dpi"] = 300


import tifffile
from scipy.ndimage import label


def rgb2gray(mask_rgb, debug=False):
    """
    # RGB to Grayscale
    ## Code logic:
    - Essentially, read in mask only image (not composite) and figure out unique tuples that represent singular RGB values.
    - A typical RBG value to produce a color value can be broken up into (R, G, B) insity values.
    - Tuples are preferable over numpy arrays as they can be easily condensed into unique elemnts using a set and then
    converted back into a list (they are hashable).
    - After uniques are found, build a dictionary with values from 1 to number of unique RGB values in the images.
    These will become the grayscale int values that we then plot.
    The resulting array should automatically upscale from 8-bit integer if needed (>256 unique elements).
    - **This only works for mask images, NOT overlays!**

    Parameters
    ----------
    mask_rgb : np.ndarray
        rgb image.
    debug : bool, optional
        Enable debugging output. The default is False.

    Returns
    -------
    result : np.ndarray
        Grayscale image.

    """

    # Iterate over every row and column in the mask image and generate a list of RGB tuples
    # There are probably more efficient ways to do this with np.reshape/np.squeeze, but for readability this was used:
    #     - if you are dealing with a larger image set, adjust as needed for a more efficient solution
    # Creating a set initially and having to check every time we add a new element for whether it is unique is inefficient

    tuple_list = list()
    for row in mask_rgb:
        for column in row:
            tuple_list.append(tuple(column))

    # Convert from list -> set -> list to only keep unique RGB values for lookup table
    # Benefit of the tuple shows here! Hashable in a set which a numpy array alone wouldn't be

    tuple_list = list(set(tuple_list))

    # Make a dictionary of RGB tuples to int index from 1 to length of unique RGB tuples
    # Add one to reserve 0 for (0,0,0) and update (0,0,0) at the end to be 0

    tuple_dict = {tuple_list[i]: i + 1 for i in range(len(tuple_list))}
    tuple_dict[(0, 0, 0)] = 0

    # Generate a numpy of colors and IDs.
    #   - The colors array will be n x 3 depending on the number of unique RGB values
    #   - The color_ids array will be n x 1 representing each of the new values for any given RGB value
    #   - This creates effectively a lookup table for converting a list of RGB values to a grayscale int equivalent

    colors = np.array(list(tuple_dict.keys()))
    color_ids = np.array(list(tuple_dict.values()))

    # Always prefer 8bit int if we can get away with it for easier viewing in windows platforms.
    #   - If there are more than 255 distinct values, then use np.uint32 as that is always positive
    #   and guarantees compatibility with even transparency based RGB codes (#AARRGGBB)

    if len(tuple_dict) < 255:
        dtype = np.uint8
    else:
        dtype = np.uint32

    # Initialize output array as the shape of the original image (minus RGB channels) and set everything to 0 initially
    # Also auto update dtype to be uint8 if possible, otherwise uint32

    result = np.zeros((mask_rgb.shape[0], mask_rgb.shape[1]), dtype=dtype)

    # Finally, check color of individual pixels in original mask against the color lookup table.
    # If all 3 RGB values match, then the index and new label of that point are returned

    label, row, column = np.where((mask_rgb == colors[:, None, None, :]).all(axis=3))

    unique_colors = np.unique(colors, axis=0)

    # Index rows and columns to now be the color ID determined above
    # Length of each of these will be mask.shape[0]*mask.shape[1], and will be automatically reshaped to fit the result array

    result[row, column] = color_ids[label]

    return result



from .dice import dice
from .h_index.h_index import h_index, h_index_single_weighted
from .hausdorff_distance import hausdorff_distance
from .jaccard import jaccard
from .percent_content_captured import percent_content_captured
from .total_error import total_error

__all__ = [
    "dice",
    "jaccard",
    "total_error",
    "percent_content_captured",
    "hausdorff_distance",
    "h_index",
    "h_index_single_weighted",
]



# Dependencies
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc

if __name__ == "__main__":

    from pathlib import Path

    import matplotlib as mpl
    import matplotlib.pylab as plt

    from cell_analysis_tools.io import load_sdt_file

    mpl.rcParams["figure.dpi"] = 600
    import numpy as np

    from cell_analysis_tools.image_processing import bin_sdt

    # variables
    laser_angular_frequency = 80e6

    ### T Cells
    # path_sdt = Path("C:/Users/Nabiki/Desktop/data/T_cells-paper/Data/011118 - Donor 4/SDT Files/Tcells-001.sdt")
    # sdt_im = load_sdt_file(path_sdt).squeeze()
    # n_rows, n_cols, n_timebins = sdt_im.shape
    # integration_time = 1 / laser_angular_frequency
    # timebins = np.linspace(0, integration_time, n_timebins, endpoint=False)
    # decay = np.sum(sdt_im, axis=(0,1))
    # plt.plot(decay)
    # plt.show()
    # plt.imshow(sdt_im.sum(axis=2))
    # plt.show()

    HERE = Path(__file__).resolve().parent
    ########### neutrohpils
    # working_dir = Path(r"C:\Users\Nabiki\Desktop\development\cell-analysis-tools\cell_analysis_tools\example_data\t_cell".replace('\\','/'))
    # path_sdt = working_dir / "Tcells-001.sdt"
    # path_sdt = Path(HERE.parent / "example_data/neutrophils/Neutrophils-021_NADH.sdt")

    ###### LOAD SDT FILE
    # Kelsey IRF's
    # irf = tifffile.imread( Path(HERE.parent / "example_data/neutrophils/Neutrophils-021_IRF.tiff"))
    # irf = read_asc("Z:/0-Projects and Experiments/KM - OMI Phasor Plots/40x_WID_2019Mar_IRF.asc")
    irf = read_asc(HERE / "example_data/irf_40xW_02_dec2017_IRF.asc")

    irf_timebins = irf[:, 0] * 1e-9  # timebins in ns
    irf_decay = irf[:, 1]  # photons count

    ###### LOAD SDT FILE
    path_sdt = Path(
        HERE / "example_data/EPC16_Day8_4n-063/LifetimeData_Cycle00001_000001.sdt"
    )

    im_sdt = load_sdt_file(path_sdt).squeeze()
    n_rows, n_cols, n_timebins = im_sdt.shape
    integration_time = 1 / laser_angular_frequency
    timebins = np.linspace(0, integration_time, n_timebins, endpoint=False)
    decay = np.sum(im_sdt, axis=(0, 1))
    plt.plot(decay)
    plt.show()
    plt.imshow(im_sdt.sum(axis=2))
    plt.show()

    # REMOVE DECAY OFFSET
    # 7x7 bin
    im_sdt_binned = bin_sdt(im_sdt, bin_size=3, debug=True)

    # threshold decays
    decays = im_sdt_binned[im_sdt_binned.sum(axis=2) > 2000]

    # calculate shift here after removing bg

    # show First 100 decays
    # for d in decays[:100]:
    #     plt.plot(d)
    # plt.show()

    # compute calibration after irf aligned
    irf_lifetime = 0
    calibration = phasor_calibration(
        laser_angular_frequency, irf_lifetime, irf_timebins, irf_decay
    )

    # COMPUTE G AND S VALUES
    array_phasor = [
        td_to_fd(laser_angular_frequency, irf_timebins, decay) for decay in decays
    ]

    # compute g and s for
    list_gs = [
        phasor_to_rectangular_point(
            ph.angle + calibration.angle, ph.magnitude * calibration.scaling_factor
        )
        for ph in array_phasor
    ]

    g = [point.g for point in list_gs]
    s = [point.s for point in list_gs]
    counts = decays.sum(axis=1)

    # plot
    figure = plt.figure()
    plt.ylim([0, 1])
    plt.xlim([0, 1])
    plt.axis("equal")
    plt.scatter(
        g, s, s=1, cmap="viridis_r", alpha=1
    )  # s=size, c= colormap_data, cmap=colormap to use
    # plt.colorbar()
    draw_universal_semicircle(figure, laser_angular_frequency)
    
    #%%
    
    from cell_analysis_tools.io import load_sdt_file
    from cell_analysis_tools.flim import (draw_universal_semicircle,
                                          phasor_to_rectangular,
                                          rectangular_to_phasor,
                                          lifetime_to_phasor
                                          )
    
    # # load irf
    irf_decay = np.loadtxt("irf.csv")   
    timebins = irf_decay[:,0]
    irf = irf_decay[:,1]
    # plt.plot(timebins,irf)
        
    # load and plot original irf
    f=0.08 #Ghz
    phasor = lifetime_to_phasor(f=f, timebins=timebins, counts=irf)
    draw_universal_semicircle(laser_angular_frequency=80e6)
    g,s = phasor_to_rectangular(phasor.angle, phasor.magnitude)
    plt.scatter(g,s)
    
    # calibate and plot original irf
    calibration = phasor_calibration(f=80e6,
                                    lifetime=0,
                                    timebins=timebins,
                                    counts=irf)    
    
    phasor_calibrated = phasor.angle + calibration.angle, phasor.magnitude * calibration.scaling_factor
    new_g, new_s = phasor_to_rectangular(phasor_calibrated[0], phasor_calibrated[1])
    plt.scatter(new_g, new_s)
    
    # plot bead data
    # load image 
    sdt = load_sdt_file("./bigger_beads_2.1ns.sdt")
    im_nadh = sdt.squeeze()
    decay = im_nadh.sum(axis=(0,1))
    
    phasor = lifetime_to_phasor(f=80e6, timebins=timebins, counts=decay)
    g,s = phasor_to_rectangular(phasor.angle, phasor.magnitude)
    plt.scatter(g, s)
    
    # calibrate 
    calibration = phasor_calibration(f=80e6,
                                            lifetime=2.2e-9,
                                            timebins=timebins,
                                            counts=decay)    
    
    phasor_calibrated = phasor.angle + calibration.angle, phasor.magnitude * calibration.scaling_factor
    new_g, new_s = phasor_to_rectangular(phasor_calibrated[0], phasor_calibrated[1])
    plt.scatter(new_g, new_s)
    
    phasor2 = rectangular_to_phasor(g, s)

    #### plot image
    plt.show()
    



# https://python.plainenglish.io/solve-graph-coloring-problem-with-greedy-algorithm-and-python-6661ab4154bd

from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt

mpl.rcParams["figure.dpi"] = 300
from collections import OrderedDict
from pprint import pprint

import numpy as np
import tifffile
from scipy.ndimage import label
from skimage.morphology import dilation, disk


def four_color_theorem(mask: np.ndarray) -> np.ndarray:
    """
    Converts a n-labeled image to an up to 4 color image.

    Parameters
    ----------
    mask : np.ndarray
        labeled mask with unique rois.

    Returns
    -------
    four_color_mask : np.ndarray
        returns a labeled image that uses up to 4 colors.
    solution_nodes : dict
        dictionary of {value : color} for each unique roi value

    """

    ### Generate Agencency Matrix
    id_nodes = list(np.unique(mask))
    id_nodes.remove(0)  # remove bg
    dict_nodes_key_roi_value = {}
    dict_nodes_key_idx = {}

    for idx, roi_value in enumerate(id_nodes):
        dict_nodes_key_roi_value[roi_value] = {}
        dict_nodes_key_roi_value[roi_value]["adj_mat_idx"] = idx
        dict_nodes_key_roi_value[roi_value]["colors"] = ["b", "c", "m", "y", 'r', 'g']
        dict_nodes_key_idx[idx] = roi_value

    adjacency_matrix = np.zeros((len(id_nodes), len(id_nodes)))

    for roi_value in dict_nodes_key_roi_value:
        pass
        roi_idx = dict_nodes_key_roi_value[roi_value]["adj_mat_idx"]
        mask_roi = np.array(mask == roi_value, dtype=np.uint16)
        # plt.imshow(mask_roi)
        # plt.show()

        # determine neighbors / degree
        # dilate and overlap with labeled mask to determine neighbors
        mask_roi_dialated = dilation(mask_roi, disk(3))
        mask_overlap = mask * mask_roi_dialated
        nodes_neighbors = list(np.unique(mask_overlap))
        nodes_neighbors.remove(roi_value)  # remove current roi
        nodes_neighbors.remove(0)  # remove bg

        #  populate adgacency matrix
        if nodes_neighbors:
            for neighbor_value in nodes_neighbors:
                neighbor_index = dict_nodes_key_roi_value[neighbor_value]["adj_mat_idx"]
                adjacency_matrix[roi_idx][neighbor_index] = 1
                adjacency_matrix[neighbor_index][roi_idx] = 1

        # dict increase degree
        dict_nodes_key_roi_value[roi_value]["degree"] = len(nodes_neighbors)

    ### Sort dictionary by largest to smallest degree
    # this will color regions with many neighbors first
    sorted_nodes = sorted(
        dict_nodes_key_roi_value.items(),
        key=lambda item: item[1]["degree"],
        reverse=True,
    )
    nodes = OrderedDict(sorted_nodes)

    # Main Algorithm
    solution_nodes = {}  # roi_value : color
    for roi_value in nodes.keys():  # iterate through roi_values
        pass
        # get list of colors in current roi, choose current color for roi
        list_roi_colors = nodes[roi_value]["colors"]
        # if roi_value == 302:
        #     print('stop')
        curr_color = list_roi_colors[0]
        solution_nodes[roi_value] = curr_color  # set solutions

        # get adjacency matrix to compare color against neighbors
        adjacent_nodes = adjacency_matrix[
            nodes[roi_value]["adj_mat_idx"]
        ]  # get adj_matrix row at this index

        # iterate through each row/neighbor in adj matrix
        for pos in range(len(adjacent_nodes)):
            pass
            # if it's a neighbor and it has the color in it's list, remove it for that neighbors list
            adj_node_idx = dict_nodes_key_idx[pos]
            if adjacent_nodes[pos] == 1 and (
                curr_color in nodes[adj_node_idx]["colors"]
            ):
                nodes[adj_node_idx]["colors"].remove(curr_color)

    # create array to store solution
    four_color_mask = np.zeros_like(mask)

    # convert from color to an intensity value for each color
    for roi_value in solution_nodes:
        pass
        # convert color to intensity
        if solution_nodes[roi_value] == "b":
            value = 50
        elif solution_nodes[roi_value] == "c":
            value = 100
        elif solution_nodes[roi_value] == "m":
            value = 150
        elif solution_nodes[roi_value] == "r":
            value = 200
        elif solution_nodes[roi_value] == "g":
            value = 225
        elif solution_nodes[roi_value] == "y":
            value = 255

        # change roi color to value
        four_color_mask[mask == roi_value] = value

    return four_color_mask, solution_nodes


if __name__ == "__main__":

    path_mask = Path("mask.tiff")
    mask = tifffile.imread(path_mask)
    four_color_mask, dict_solution = four_color_theorem(mask)

    ## Display image
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))

    ax[0].set_axis_off()
    ax[0].set_title(f"original mask \n unique rois: {len(np.unique(mask))-1}")
    ax[0].imshow(mask)

    ax[1].set_axis_off()
    ax[1].set_title(
        f"four color mask \n unique rois: {len(np.unique(four_color_mask))-1}"
    )
    ax[1].imshow(four_color_mask)

    plt.imshow(four_color_mask)
    plt.show()



# Dependencies
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


# SHIFT IRF
def estimate_and_shift_irf(decay, irf_decay, debug=False):
    """
    Estimates shift  given a decay and an IRF. 

    Parameters
    -----------
        decay : ndarray
            1D array containing decay curve
        irf_decay : ndarray
            1D array containing IRF  

        debug : bool
            Optional flag to display intermediate calculations

    Returns
    -------
        irf_decay_shifted : ndarray
            Shifted IRF as a 1d array
        shift : int
            value IRF was shifted by

    Note
    ----
        IRF and decay should NOT have low SNR or gradient function will produce incorrect alignment   
    
    
    .. image:: ./resources/flim_estimate_and_shift_irf.png
        :width: 800
        :alt: Image of the estimated placement of IRF for a decay
        
    """

    # TODO: smooth signals
    # TODO: center IRF on middle of rising decay gradient or a little before
    # peak should be near the middle of decay rise or closer to the left side
    # account for offset of decay vertically
    num_timebins = len(decay)
    # SDT DECAY
    decay_grad = np.gradient(decay)
    decay_rising = decay_grad.copy()
    decay_rising[decay_grad < 0] = 0  # take all positive gradient
    if debug:
        plt.title("Decay Gradient")
        plt.plot(decay)
        plt.plot(decay_grad)
        plt.plot(decay_rising)
        plt.legend(["decay", "gradient", "positive gradient"])
        plt.show()

    # IRF --> Extract peak of rising
    irf_decay_grad = np.gradient(irf_decay)

    irf_rising = irf_decay_grad.copy()
    irf_rising[irf_rising < 0] = 0
    if debug:
        plt.title("IRF Gradient")
        plt.plot(irf_decay)
        plt.plot(irf_decay_grad)
        plt.plot(irf_rising)
        plt.legend(["decay", "gradient", "positive gradient"])
        plt.show()

    # compute shift
    correlated = np.correlate(decay_rising, irf_rising, mode="full")
    peak_correlated = np.argmax(correlated)
    shift = peak_correlated - len(
        decay_rising
    )  # because length is (len_d1 + len_d2 - 1)

    irf_decay_shifted = np.roll(irf_decay, shift)

    # if debug: #visualize shifted correlation
    #     plt.plot(normalize(decay_rising))
    #     plt.plot(normalize(irf_rising))
    #     plt.plot(normalize(correlated))
    #     plt.plot(np.roll(normalize(correlated), -np.argmax(correlated) ))
    #     plt.legend(["decay rising gradient","irf rising gradient","correlated signal"])
    #     plt.show()

    if debug:  # visualize
        plt.title(f"Shift estimation \n{shift} ps")
        plt.xlabel("Time(ps)")
        plt.ylabel("Counts")
        plt.plot(normalize(decay))
        plt.plot(normalize(irf_decay))
        plt.plot(normalize(irf_decay_shifted))
        plt.legend(["decay", "irf", "shifted irf"])
        plt.show()

    return irf_decay_shifted, shift


if __name__ == "__main__":
    pass

    from cell_analysis_tools.io import load_sdt_file
    
    # load irf
    irf_decay = np.loadtxt("irf.csv")    
    plt.plot(irf_decay[:,0],irf_decay[:,1])
    
    # load image 
    sdt = load_sdt_file("./resources/test_image.sdt")
    im_nadh = sdt[1,...]
    # plt.imshow(im_nadh.sum(axis=2))
    # plt.show()
    
    decay = im_nadh.sum(axis=(0,1))
    decay = np.roll(decay,10)
    # plt.plot(decay)
    # plt.show()
    
    # visualize and plot IRF
    irf_decay_shifted, shift = estimate_and_shift_irf(decay, irf_decay[:,1], debug=True)
    
    
    
    
    
    


import numpy as np
from numpy.ma import masked_array


def intensity_sum(roi, im_itensity):
    """The sum of the pixel intensities within the segmented region.
    """

    roi_inverted = np.invert(np.array(roi).astype(bool))
    roi_masked = masked_array(im_itensity, mask=roi_inverted)

    return np.sum(roi_masked, axis=(0, 1))



import numpy as np

from .helper import _validate_array_and_make_bool


# https://gist.github.com/brunodoamaral/e130b4e97aa4ebc468225b7ce39b3137
def dice(im1, im2, empty_score=1.0):
    """
    Computes the dice coefficient/F1 score, a measure of average similarity.

    Parameters
    ----------
    im1 : array-like, bool
        Any array of arbitrary size. If not boolean, will be converted.
    im2 : array-like, bool
        Any other array of identical size. If not boolean, will be converted.
    empty-score : int
        Any other array of identical size. If not boolean, will be converted.
    Returns
    -------
    float
        Dice coefficient as a float on range [0,1]. \n
        Maximum similarity = 1 \n
        No similarity = 0 \n
        Both are empty (sum eq to zero) = empty_score
        
    Note
    -----
    The order of inputs for `dice` is irrelevant. The result will be
    identical if `im1` and `im2` are switched.
    """
    im1 = _validate_array_and_make_bool(im1)
    im2 = _validate_array_and_make_bool(im2)

    if im1.shape != im2.shape:
        raise ValueError("Shape mismatch: im1 and im2 must have the same shape.")

    im_sum = im1.sum() + im2.sum()
    if im_sum == 0:  # no true values in either image
        print("warning: no true values in either array, returning empty_score=1")
        return empty_score

    # Compute Dice coefficient
    intersection = np.logical_and(im1, im2)

    return 2.0 * intersection.sum() / im_sum



def magnitude(roi):
    """The magnitude of a vector, which can be computed at either a pixel or mitochondrion level.

    """
    raise NotImplementedError


def angle(roi):
    """The angle of a vector, which can be computed at either a pixel or mitochondrion level.

    """
    raise NotImplementedError


def dx(roi):
    """The x component of a motion vector, which can be computed at either a pixel or mitochondrion level.

    """
    raise NotImplementedError


def dy(roi):
    """The y component of a motion vector, which can be computed at either a pixel or mitochondrion level.

    """
    raise NotImplementedError


def median_std_intensity(roi):
    """The median of standard deviation of intensity values in a segmented region. These values are obtained by computing the standard deviation of the intensity values of a pixel across all frames of a video.

    """
    raise NotImplementedError



from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt

mpl.rcParams["figure.dpi"] = 300
from collections import OrderedDict
from pprint import pprint

import numpy as np
import tifffile
from scipy.ndimage import label
from skimage.measure import regionprops
from skimage.morphology import binary_closing, dilation, disk, remove_small_objects


def four_color_to_unique(mask: np.ndarray, debug: bool = False) -> np.ndarray:
    """
    Converts an n-label image into a 

    Parameters
    ----------
    mask : np.ndarray
        labeled mask.
    debug : bool, optional
        show output re-labeled mask. The default is False.

    Returns
    -------
    output_mask : np.ndarray
        Mask relabeled to have unique roi values for each connected component

    """
    # determine unique labels and exclude bg(0)
    list_unique_values = list(np.unique(mask))
    list_unique_values.remove(0)  # remove bg

    # label counter and placeholder array
    roi_counter = 1
    inter_mask = np.zeros_like(mask)

    # iterate through 4 color labels
    for value in list_unique_values:
        pass
        temp_mask = np.array(mask == value, dtype=np.uint16)

        # TODO exclude these next lines, specific to Gina's masks!
        labels, _ = label(temp_mask)

        unique_labels = list(np.unique(labels))
        unique_labels.remove(0)
        for roi_value in unique_labels:
            inter_mask[labels == roi_value] = roi_counter
            roi_counter += 1  # increment roi_value

        # compute regionprops
        # sort by centroid
        props = regionprops(inter_mask)
        dict_value_location = OrderedDict({roi.label: roi.centroid for roi in props})

        # sort dict by tuple (rows, cols)
        dict_value_location_sorted = sorted(
            dict_value_location.items(), key=lambda item: (item[1][0], item[1][1])
        )

        # reoder roi values in increasing order
        roi_counter = 1
        output_mask = np.zeros_like(inter_mask)
        for roi_val, _ in dict_value_location_sorted:
            pass
            output_mask[inter_mask == roi_val] = roi_counter
            roi_counter += 1

    if debug:
        plt.imshow(output_mask)
        plt.show()

    return output_mask


if __name__ == "__main__":

    from four_color_theorem.four_colors import four_color_theorem

    mask = tifffile.imread("mask.tiff")

    # plt.imshow(mask)
    mask_fc, _ = four_color_theorem(mask)
    mask_unique = four_color_to_unique(mask_fc)

    # temp_mask = np.array(binary_closing(temp_mask, disk(1)), dtype=np.uint16)
    labeled_mask = remove_small_objects(mask_unique, min_size=30)
    plt.title(f"unique values \n unique rois: {len(np.unique(labeled_mask))-1}")
    plt.imshow(labeled_mask)
    plt.show()



# Dependencies
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve
import tifffile


from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc

def bin_image(image, bin_factor):
    """
    This function takes in an lifetime image and bins 
    the decays of its histogram given a bin factor. A bin of 2 will reduce
    the size of the image by 2. It will do so by forming a kernel of 
    bin_factor x bin_factor and sum of those decays into a resulting pixel
    
    Parameters
    ----------

    image : ndarray
        image to bin, must be a 3d array of shape (x,y,t)
    bin_factor : int
        pixel radius(including diagonals) of bins to create. Must be EVEN.
    
    Returns
    -------
        binned_image : ndarray
            new image with binned pixels
            
            
    .. image:: ./resources/flim_bin_image.png 
        :width: 400
        :alt: before and after binning function
            
            
    """

    #    image = image_thresholded
    #    bin_factor = 9
    
    assert bin_factor % 2 ==0, "Error: Bin factor must be even."

    # GET IMAGE DIMENSIONS
    x, y, num_timebins = image.shape
    axis_x, axis_y, axis_timebins = (0, 1, 2)
    # tif.imshow(np.sum(image,axis=axis_timebins))

    # new image dimensions
    # If not divisible by downsampling factor:
    # (y - remainder + downsampling_factor_for_padding)/downsampling factor
    # y-remainder ==> makes it divisible by downsampling factor
    # + bin_factor padds it to make it bigger, then divide
    remainder_x = x % bin_factor
    remainder_y = y % bin_factor
    subsampled_x = int(
        x / bin_factor
        if x % bin_factor == 0
        else (x + bin_factor - remainder_x) / bin_factor
    )
    subsampled_y = int(
        y / bin_factor
        if y % bin_factor == 0
        else (y + bin_factor - remainder_y) / bin_factor
    )

    """ need to pad with zeros if not divisible by bin_factor"""
    padded_x = int(subsampled_x * bin_factor)
    padded_y = int(subsampled_y * bin_factor)

    # Original image padded with zeros
    image_padded = image.copy()
    # tif.imshow(np.sum(image_padded,axis=0))

    # pad image dimensions
    col_to_add = padded_x - x
    pad_right = int(np.floor(col_to_add / 2) + padded_x % x)  # pad half pixels added
    pad_left = int(np.floor(col_to_add / 2))

    row_to_add = padded_y - y
    pad_top = int(np.floor(row_to_add / 2) + padded_y % y)
    pad_bottom = int(np.floor(row_to_add / 2))

    # np.pad((time),(rows/yaxis),(columns/xaxis))
    image_padded = np.pad(
        image_padded,
        ((0, 0), (pad_left, pad_right), (pad_top, pad_bottom)),
        "constant",
        constant_values=0,
    )

    sub_matrices = np.zeros(
        (subsampled_x, subsampled_y, num_timebins, bin_factor ** 2), dtype=object
    )

    # variable to store binned image
    binned_image = np.zeros((subsampled_x, subsampled_y, num_timebins))

    index = 0
    # GENERATE bin_factor^2 NUMBER OF SUBMATRICES
    for column in range(bin_factor):
        for row in range(bin_factor):
            temp_matrix = image_padded[
                column:padded_x:bin_factor, row:padded_y:bin_factor, :
            ]
            #            tif.imshow(np.sum(temp_matrix, axis=0))

            # add matrix to array
            sub_matrices[:, :, :, index] = temp_matrix
            index += 1

            # display sub_matrices
            # temp = sub_matrices[:,:,:,index] # last dimension is sub matrix
            # temp2 = temp.astype(float) # cast from object to float
            # plt.imshow(np.sum(temp2, axis=0)) # show submatrix

            # keep adding to the image
            binned_image = binned_image + temp_matrix

        """ improvements
        * preserve resolution like SPCImage, using kernel size
        * fft of original image
        * fft of kernel, padded to equal image size
        * take multiplication
        * convert back to time domain
        
        can we assume image is a square?
        """
        # tif.imshow(np.sum(binned_image, axis=0))

    return binned_image

if __name__ == "__main__":
    

    import matplotlib as mpl
    mpl.rcParams['figure.dpi'] = 300
    
    default_rng = np.random.default_rng(seed=1)
    
    
    # generate temporary image
    im = default_rng.random((256,256,256))
    bin_factor = 4
    im_binned = bin_image(im, bin_factor)
    
    fig, ax = plt.subplots(1,2, figsize=(5,3))
    
    fig.suptitle("bin image")
    ax[0].imshow(im.sum(axis=2))
    ax[0].set_axis_off()
    ax[0].set_title(f"original \n{im.shape}")
    
    ax[1].imshow(im_binned.sum(axis=2))
    ax[1].set_axis_off()
    ax[1].set_title(f"bin of {bin_factor} \n{im_binned.shape}")
    
    plt.savefig("./resources/fig_bin_image.png")
    
    plt.show()
    
    
    
    
    
    
    
    
    
    
    
    
        




__version__ = "0.0.14"

from . import flim
from . import image_processing
from . import io
from . import metrics
from . import visualization


__all__ = [
    'flim',
    'image_processing',
    'io',
    'metrics',
    'visualization',
    ]


def normalize(im):
    """
    Normalizes an image to 0 and 1 by subtrating min value and 
    dividing by new max value.
    
    Parameters
    ----------
    im : array-like
        intensity image.

    Returns
    -------
    array-like
        array normalized to 0 and 1.

    """
    return (im - im.min()) / (im.max() - im.min())



import zipfile

import numpy
import numpy as np
import tifffile
from matplotlib import pyplot as plt
from read_roi import read_roi_zip
from skimage.draw import polygon2mask


def load_sdt_file(file_path):
    """ 
    Loads an sdt file and attemps to reshape the data into a cube corresponding to common file shapes.
    
    
    Parameters
    -----------
        
    file_path : pathlib path
         Path path to the sdt file
    
    Returns
    -------
    
    image : np.ndarray
        A 3d-array representation of image.
    """

    with zipfile.ZipFile(file_path) as myzip:
        z1 = myzip.infolist()[
            0
        ]  # "data_block" or sdt bruker uses "data_block001" for multi-sdt"
        with myzip.open(z1.filename) as myfile:
            data = myfile.read()
            data = np.frombuffer(data, np.uint16)

    # format == [channel, x, y, num_timebins]

    # todo
    """  figure out how to extract this info from sdt file"""
    img_1_256_256_256 = 16777216
    img_1_512_512_256 = 67108864
    img_2_512_512_256 = 134217728  # array size
    img_3_512_512_256 = 201326592
    img_2_256_256_256 = 2 * 256 * 256 * 256
    if len(data) == img_2_512_512_256:
        c, x, y, z = (2, 512, 512, 256)
    if len(data) == img_1_256_256_256:
        c, x, y, z = (1, 256, 256, 256)
    if len(data) == img_1_512_512_256:
        c, x, y, z = (1, 512, 512, 256)
    if len(data) == img_3_512_512_256:
        c, x, y, z = (3, 512, 512, 256)
    if len(data) == img_2_256_256_256:
        c, x, y, z = (2, 256, 256, 256)

        # if c == 1:
        #     ''' order is [XYT] '''
        #     numpy_image = np.reshape(data, (x, y, z))
        #     # if debug: tif.imshow(np.sum(image,axis=2))
        #     return numpy_image

        # else:
        """ order is [CXYT] """
    numpy_image = np.reshape(data, (c, x, y, z))
    return np.float32(numpy_image)


def load_sdt_data(filepath):
    """
    Loads the data of an SDT file, reshaping the output is necessary as this
    outputs a 1d array.

    .. code-block::python
    
        An image of length 16777216 should be reshaped to (1, 256,256,256)
        >>> image = np.reshape(data, (1, 256, 256, 256))
    
        ### 2 Channel
        An image of length 33554432 should be reshaped to (2 ch, 256,256,256)
        >>> image = np.reshape(data, (2, 256, 256, 256))

    Parameters
    ----------
    filepath : string, pathlib path
        path to the input image

    Returns
    -------
    data : np.ndarray
        1D array containing all channel and pixel data

    """
    with zipfile.ZipFile(filepath) as myzip:
        z1 = myzip.infolist()[
            0
        ]  # "data_block" or sdt bruker uses "data_block001" for multi-sdt"
        with myzip.open(z1.filename) as myfile:
            data = myfile.read()
            data = np.frombuffer(data, np.uint16)

    return data



# Dependencies
import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def phasor_to_rectangular(angle, magnitude):
    """ Gets x and y position of phasor in rectangular coordinates

    Parameters
    ----------
        angle  : float 
            phasor angle in radians
        magnitude  : float 
            phasor magnitude
    Returns
    -------
        g  : float 
            x axis coordinate
        s  : float
            y axis coordinate
    """
    point_rect = coll.namedtuple("point_rect", "g s")
    g = magnitude * np.cos(angle)
    s = magnitude * np.sin(angle)
    return point_rect(g=g, s=s)



from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np

from cell_analysis_tools.io import load_sdt_file

mpl.rcParams["figure.dpi"] = 300


def bin_3d(im, bin_size, stride=1, pad_value=0, debug=False):
    """
    Bins a 3d array along 3rd dimension, usually timebins in flim images (x,y,t).

    Parameters
    ----------
    im : ndarray
        3d ndarray containing FLIM data (x,y,t).
    bin_size : int
        number of pixels to look before and after center pixel.
        kernel = bin_size + 1 + bin_size
    stride : int, optional
        Number of pixels to move when raster scanning the image. The default is 1.
    pad_value : int, optional
        value to pad the image with. The default is 0.
    debug : bool, optional
        Show debugging output. The default is False.

    Returns
    -------
    im_binned : ndarray
        binned array of same size as the input

    """

    if stride != 1:
        print("strides larger than 1 not yet implemented")

    kernel_length = bin_size * 2 + 1

    rows_before = bin_size
    rows_after = bin_size
    cols_before = bin_size
    cols_after = bin_size

    n_rows, n_cols, _ = im.shape

    padded_sdt = np.pad(
        im,  # this takes in width(cols) and height(rows)
        (
            (rows_before, rows_after),
            (cols_before, cols_after),
            (0, 0),  # dont pad 3rd dim
        ),
        "constant",
        constant_values=pad_value,
    )

    im_binned = np.zeros(im.shape)

    for row_idx in np.arange(n_rows):
        for col_idx in np.arange(n_cols):

            # run this and see if this works!
            # starts at 0(row_idx) : kernel_length + 0
            # then 1 (row_idx) : kernel_length + 1 ...etc
            # up to length n-2
            conv_decay = padded_sdt[
                row_idx : kernel_length + row_idx, col_idx : kernel_length + col_idx, :
            ]
            im_binned[row_idx, col_idx, :] = conv_decay.sum(axis=(0, 1))

    if debug:
        plt.title(f"kernel: {kernel_length}x{kernel_length}")
        plt.imshow(im_binned.sum(axis=2))
        plt.show()

    return im_binned


if __name__ == "__main__":

    ## bin_3d test
    HERE = Path(__file__).resolve().parent
    data = HERE.parent / r"example_data\t_cell\Tcells-001.sdt".replace("\\", "/")

    # load SDT
    im = load_sdt_file(data)
    for ch in np.arange(im.shape[0]):
        plt.title(f"channel: {ch}")
        plt.imshow(im[ch, ...].sum(axis=2))
        plt.show()

    # select channel with data
    im = im[1, ...]

    # run with debug on
    bin_3d(im, bin_size=3, debug=True)



import matplotlib.pylab as plt
import numpy as np
from sklearn.cluster import KMeans

from .normalize import normalize


def kmeans_threshold(im, k, n_brightest_clusters, show_image=False):
    """
    Given an image, this function will apply k-means clustering and return the mask that 
    includes the n brightest clusters
    Parameters
    ----------
    im : array-like
        intensity image.
    k : int, 
        number of clusters for k means algorithm
    n_brightest_clusters : int, 
        number of brightest clusers to keep. Must be < k.
        
    show_image : bool, optional
        When debugging,this will display the original image next to the k means thresholded image. The default is False.

    Returns
    -------
    mask : int
        binary mask

    """
    # k = 3
    # n_brightest_clusters = 1 # this should be < k

    if n_brightest_clusters >= k:
        print("n_brightest_clusters must be < k")
        return

    # normalize to 0 and 1
    # im = (im - im.min()) / (im.max() - im.min())  # subtract baseline / new_max value)
    im = normalize(im)

    rows, cols = im.shape
    X = im.reshape((rows * cols, 1))  # reshape to (n_samples, n_features)

    # apply k means to image
    kmeans = KMeans(n_clusters=k, random_state=0).fit(X)

    #### keep brightest clusters by zeroing out dimmestclusters
    num_blanked_dimmer_clusters = (
        k - n_brightest_clusters
    )  # calculate n_dimm clusters based on clusters to keep
    # get list of clusters, first n are dimmest/smallest
    # https://stackoverflow.com/questions/6910641/how-do-i-get-indices-of-n-maximum-values-in-a-numpy-array
    # https://stackoverflow.com/questions/34226400/find-the-index-of-the-k-smallest-values-of-a-numpy-array
    n_dimmest_clusters = np.argpartition(
        kmeans.cluster_centers_[:, 0], num_blanked_dimmer_clusters
    )
    indices_smallest = n_dimmest_clusters[
        :num_blanked_dimmer_clusters
    ]  # get a hold of n dimmest clusters

    # make dimmest clusters == 0
    newLUT = kmeans.cluster_centers_.copy()
    for x in indices_smallest:
        newLUT[x, :] = 0

    # replace labels with cluster values
    clustered_im = newLUT[kmeans.labels_]
    ####

    # Reshape back the image from 2D to 3D image
    clustered_imaged = clustered_im.reshape(rows, cols)
    mask = (clustered_imaged > 0).astype(int)  # make a binary mask

    if show_image:
        plt.title(f"num clusters (k): {k} , clusters kept: {n_brightest_clusters}")
        divider = np.ones((rows, 5))
        plt.imshow(np.c_[im, divider, mask])
        plt.show()

    return mask



import numpy as np
import pandas as pd
import umap

#%% UMAP


def compute_umap(data_values: np.ndarray, **kwargs) -> pd.DataFrame:
    """
        Parameters
        ----------
        data_values : np.ndarray
            array of values to input into the reducer. Rows should be roi values,
            columns are features
        random_state : int, optional
            allows consistent initialization of reducer. The default is 0.
        **kwargs : dict
            additional parameters can be passed in here for the reducer.
        
        Returns
        -------
        df_umap : pd.DataFrame
            DataFrame containing umap embeddings.
        reducer : Object
            umap reducer object.
            
        .. note::
            See the umap-learn documentation for default and additional parameters
            `https://umap-learn.readthedocs.io/en/latest/api.html <https://umap-learn.readthedocs.io/en/latest/api.html>`_
            
        
        """

    reducer = umap.UMAP(random_state=0, **kwargs)
    fit_umap = reducer.fit_transform(data_values)
    df_umap = pd.DataFrame(fit_umap, columns=["umap_x", "umap_y"])
    return df_umap, reducer


#%% PCA

from sklearn.decomposition import PCA


def compute_pca(
    data_values: np.ndarray, n_components: int = 2, **kwargs
) -> pd.DataFrame:
    """
    Computes PCA given input data

    Parameters
    ----------
    data_values : np.ndarray
        array of values to input into the reducer. Rows should be roi values,
        columns are features
    n_components : int, optional
        number of components for the PCA. The default is 2.

    Returns
    -------
    df_pca : pd.DataFrame
        Dataframe containing the components of the PCA.
    pca : Object
        PCA reducer object.

    """
    pca = PCA(n_components=n_components, **kwargs)
    principal_components = pca.fit_transform(data_values)
    df_pca = pd.DataFrame(
        data=principal_components,
        columns=["principal component 1", "principal component 2"],
    )
    return df_pca, pca


#%% tSNE

from sklearn import manifold


def compute_tsne(
    data_values: np.ndarray, random_state: int = 0, **kwargs
) -> np.ndarray:

    """
    Computes t-distributed stochastic neighbor embedding. 

    Parameters
    ----------
    data_values : np.ndarray
        array of values to input into the reducer. Rows should be roi values,
        columns are features.
    random_state : int, optional
        allows consistent initialization of reducer. The default is 0.

    Returns
    -------
    df_tsne : pd.DataFrame
        Dataframe containing the embeddings of tsne.
    tsne : Object
        tsne reducer object.

    """

    tsne = manifold.TSNE(random_state=random_state, **kwargs)
    principal_components = tsne.fit_transform(data_values)
    df_tsne = pd.DataFrame(data=principal_components, columns=["tsne_1", "tsne_2"])

    return df_tsne, tsne


#%% plot data

import matplotlib as mpl

# plotting
import matplotlib.pylab as plt

mpl.rcParams["figure.dpi"] = 300


def plot_data(data, labels, s=3, title=""):

    data["labels"] = labels
    dict_classes = {0: "class 1", 1: "class 2"}

    # dict_colors = { 0 : 'b', 1 : 'g'}

    for label in np.unique(data["labels"]):
        pass
        df_plot = data[data["labels"] == label]
        x = df_plot.iloc[:, 0]
        y = df_plot.iloc[:, 1]
        scatter = plt.scatter(
            x,
            y,
            # c=list_colors,
            label=dict_classes[label],
            s=s,
        )
    plt.title(title)
    plt.xlabel(x.name)
    plt.ylabel(y.name)
    plt.legend()
    plt.show()


#%%
if __name__ == "__main__":
    #%%
    from sklearn.datasets import make_classification
    from sklearn.preprocessing import StandardScaler

    x, y = make_classification(
        n_classes=2,
        n_features=10,  # e.g. omi parameters
        n_samples=1000,
        random_state=0,
    )

    list_features = [
        # 'nadh_intensity_mean',
        "nadh_a1_mean",
        "nadh_a2_mean",
        "nadh_t1_mean",
        "nadh_t2_mean",
        "nadh_tau_mean_mean",
        # 'fad_intensity_mean',
        "fad_a1_mean",
        "fad_a2_mean",
        "fad_t1_mean",
        "fad_t2_mean",
        "fad_tau_mean_mean",
        # 'redox_ratio_mean'
    ]

    df_data = pd.DataFrame(x, columns=list_features)

    scaled_data = StandardScaler().fit_transform(x)

    # umap
    df_umap, reducer_umap = compute_umap(scaled_data, metric="manhattan")
    plot_data(df_umap, labels=y, title="UMAP")

    # tsne
    df_tsne, reducer_tsne = compute_tsne(scaled_data)
    plot_data(df_tsne, labels=y, title="t-SNE")

    # pca
    df_pca, reducer_pca = compute_pca(scaled_data)
    plot_data(df_pca, labels=y, title="Principal Component Analysis")



from scipy.spatial.distance import directed_hausdorff

from .helper import _validate_array_and_make_bool


def hausdorff_distance(mask_pred, mask_gt):
    """
    Calculates the Hausdorff distance for a given image, provided a ground truth. Both arrays must have the same number of columns. 
    see https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.directed_hausdorff.html
    Parameters
    ----------
    mask_pred : np.ndarray
        predicted segmentation mask
    mask_gt : np.ndarray
        ground truth mask

    Returns
    -------
        ddouble
            The directed Hausdorff distance between arrays u and v,
        index_1int
            index of point contributing to Hausdorff pair in u
        index_2int
            index of point contributing to Hausdorff pair in v

    """

    mask_pred = _validate_array_and_make_bool(mask_pred)
    mask_gt = _validate_array_and_make_bool(mask_gt)

    return directed_hausdorff(mask_pred, mask_gt)



import numpy as np

from .helper import _validate_array_and_make_bool


def total_error(mask_gt, mask_pred, weight_fn=1, weight_fp=1):
    """
    Computes the percent of misclassified pixels on an image. 
    In the case where one is more important than the 
    other, a weighted average may be used: c0FP + c1FN
    
    'https://stats.stackexchange.com/questions/273537/f1-dice-score-vs-iou <https://stats.stackexchange.com/questions/273537/f1-dice-score-vs-iou>'_

    
    Parameters
    ----------
    mask_gt : bool ndarray
        groudn truth mask.
    mask_predicted : bool ndarray
        DESCRIPTION.
    weight_fn : int, optional
        scalar weight applied to false negatives. The default is 1.
    weight_fp : int, optional
        scalar weight applied to false positives. The default is 1.

    Returns
    -------
    percent: float
        total error of the predicted mask given a ground truth mask.

    """

    # convert to bool
    mask_gt = _validate_array_and_make_bool(mask_gt)
    mask_pred = _validate_array_and_make_bool(mask_pred)

    # subtraction operation requires arrays to be int
    # > 0 then convers then back to bool
    fp = (mask_pred.astype(int) - mask_gt.astype(int)) > 0  # bool
    fn = (mask_gt.astype(int) - mask_pred.astype(int)) > 0  # bool

    n_rows, n_cols = mask_gt.shape
    # return (np.sum(fn)*weight_fn + np.sum(fp)*weight_fp) / (n_rows * n_cols )
    return (np.sum(fn) * weight_fn + np.sum(fp) * weight_fp) / (
        (n_rows * n_cols) * ((weight_fp + weight_fn) / 2)
    )



from pathlib import Path
import re
import tifffile
import matplotlib.pylab as plt
import numpy as np

from skimage.measure import regionprops
from skimage.morphology import closing, disk, remove_small_objects, label
from cell_analysis_tools.visualization import compare_images
from tqdm import tqdm
import tifffile

import matplotlib as mpl
mpl.rcParams['figure.dpi'] = 300
 #%%

def remove_small_areas_fill_regions(mask : np.array,
                                    region_min_size : int = 100,
                                    footprint_area_closing : int = 2,
                                    debug=False):
    """
    Function was created to clean up masks that may have stray pixels or regions. 

    Parameters
    ----------
    mask : np.array
        original mask.
    region_min_size : int, optional
        minimum size of connected component to be removed. The default is 100.
    footprint_area_closing : int, optional
        When merging images radius of disk to use. The default is 2.
    debug : TYPE, optional
        Enable/Disable display of intermediate images for debugging function. The default is False.

    Returns
    -------
    np.array of revised mask

    """    

    if debug:
        plt.title("input mask")
        plt.imshow(mask)
        plt.show()
    
    mask_no_small_objects = remove_small_objects(mask, region_min_size)

    if debug:
        plt.title("after removed small objects ")
        plt.imshow(mask_no_small_objects)
        plt.show()
     
    mask_revised = np.zeros_like(mask)
    for label_value in np.unique(mask_no_small_objects):
        pass
    
        # isolate roi
        mask_roi = mask == label_value
        
        # get largest region
        mask_roi_labels_mask = label(mask_roi)
        roi = sorted(regionprops(mask_roi_labels_mask), key=lambda r : r.area, reverse=True)[0]
        mask_largest = mask_roi_labels_mask == roi.label

        # fill holes in roi
        mask_closing = closing(mask_largest,footprint=disk(footprint_area_closing))
        mask_revised[mask_closing] = label_value
    
    if debug:
        plt.title("final output")
        plt.imshow(mask_revised)
        plt.show()

    return mask_revised

if __name__== "__main__":
    
    path_mask = Path(r"/mnt/Z/0-Projects and Experiments/TQ - cardiomyocyte maturation/datasets/H9/DAY 90/masks/H9_DAY_90_2n_photons_mask_cell.tiff")

    mask = tifffile.imread(path_mask)
    plt.imshow(mask)
    plt.show()
    
    mask_revised = remove_small_areas_fill_regions(mask,
                                                   region_min_size = 10,
                                                   footprint_area_closing=10,
                                                   debug=True)
    plt.imshow(mask_revised)
    plt.show()





import numpy as np
from .helper import _validate_array_and_make_bool


def percent_content_captured(mask_a, mask_b):
    """
    Computes the content of mask_a captured in mask_b and returns it as a percent. (A intersect B)/A
    # From the paper `Mitohacker <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7642274/>`

    Parameters
    ----------
    mask_a : ndarray
        boolean mask.
    mask_b : ndarray
        boolean mask.

    Returns
    -------
    percent: float
        percent of mask_a content captured by mask_b.

    """
    mask_a = _validate_array_and_make_bool(mask_a)
    mask_b = _validate_array_and_make_bool(mask_b)

    return np.sum(np.logical_and(mask_a, mask_b)) / np.sum(mask_a)



# Dependencies
import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def rectangular_to_phasor_lifetimes_array(g, s):
    """ 
    Takes an array(image) of g and s points and
    converts them to angle and magnitude phasor arrays

    Parameters
    ----------
        g  : float 
            array of g coordinates (x-coordinates)
        s  : float 
            array of s coordinates (y-coordinates)

    Returns
    -------
        PhasorArray object : float
            lifetime_angles_array - array of angles for each pixel
            lifetime_magnitudes_array - array of magnitudes for each pixel
    """
    PhasorLifetimesArray = coll.namedtuple("PhasorLifetimesArray", "angles magnitudes")

    # calculate angles and magnitudes for all points
    angles = np.pi - np.arctan2(s, -g)
    magnitudes = np.sqrt(g ** 2 + s ** 2)
    return PhasorLifetimesArray(angles=angles, magnitudes=magnitudes)


if __name__ == "__main__":
    
    from cell_analysis_tools.io import load_sdt_file

    # load image 
    sdt = load_sdt_file("./resources/test_image.sdt")
    im_nadh = sdt[1,...]
    plt.imshow(im_nadh.sum(axis=2))
    plt.show()
    
    


import matplotlib as mpl
import matplotlib.pylab as plt

mpl.rcParams["figure.dpi"] == 300
import numpy as np
from skimage.color import label2rgb

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.metrics import dice, total_error


def image_show(image):
    """

    Parameters
    ----------
    image : ndarray
        image to show.

    Returns
    -------
    None.

    """
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 10))
    ax.imshow(image)  # , cmap='gray'
    ax.axis("off")
    plt.show()
    return fig, ax


def compare_images(title1, im1, title2, im2, suptitle=None, figsize=(10, 5), save_path=None) -> None:
    """
     
        Parameters
        ----------
        im1 : np.ndarray
            image 1.
        title1 : str
            title for image 1.
        im2 : np.ndarray
            image 2.
        title2 : str
            title for image 2.
        figsize : TYPE, optional
            size of figure. The default is (10, 5).
    
        Returns
        -------
        None.
    
    """
    fig, ax = plt.subplots(1, 2, figsize=figsize)
    if suptitle:
        fig.suptitle(suptitle)
    ax[0].title.set_text(title1)
    ax[0].imshow(im1)
    ax[0].set_axis_off()

    ax[1].title.set_text(title2)
    ax[1].imshow(im2)
    ax[1].set_axis_off()
    if save_path:
        plt.savefig(save_path, bbox_inches='tight')
    plt.show()


def compare_orig_mask_gt_pred(
    im: np.ndarray, mask_gt: np.ndarray, mask_pred: np.ndarray, title: str = ""
) -> None:
    """
    Simple function for comparing oringal image and ground truth image


    Parameters
    ----------
    im : np.ndarray
        original image.
    mask_gt : np.ndarray
        ground truth image.
    mask_pred : np.ndarray
        predicted mask.
    title : str, optional
        title of the plot, usually the origina filename. The default is "".

    Returns
    -------
    None
        function only just plots data.

    """
    alpha = 0.5
    im_overlay = label2rgb(
        mask_pred, normalize(im), bg_label=0, alpha=alpha, image_alpha=1, kind="overlay"
    )

    fig, ax = plt.subplots(2, 3, figsize=(10, 7))

    plt.suptitle(title)
    ax[0, 0].title.set_text(f"original")
    ax[0, 0].set_axis_off()
    ax[0, 0].imshow(im)

    # overlayed
    dice_coeff = dice(mask_pred, mask_gt)
    ax[0, 1].title.set_text(f"overlayed mask_pred")
    ax[0, 1].set_axis_off()
    ax[0, 1].imshow(im_overlay)

    # mask gt
    ax[1, 0].title.set_text(f"mask_gt")
    ax[1, 0].set_axis_off()
    ax[1, 0].imshow(mask_gt)

    # mask pred
    ax[1, 1].title.set_text(f"mask_pred \n dice: {dice_coeff:.4f}")
    ax[1, 1].set_axis_off()
    ax[1, 1].imshow(mask_pred)

    ## XOR
    mask_xor = np.logical_xor(mask_gt, mask_pred)

    error_total = total_error(mask_gt, mask_pred)
    ax[0, 2].title.set_text(f"mask_xor\n total error: {(error_total*100):.3f}")
    ax[0, 2].set_axis_off()
    ax[0, 2].imshow(mask_xor)

    ax[1, 2].set_axis_off()
    plt.show()


if __name__ == "__main__":

    import numpy as np

    im = np.random.rand(40, 40)
    compare_orig_mask_gt_pred(im, im, im)

    # print("TODO add test code")

    # im_orig = np.random.rand(512,512)
    # im_gt = np.round(im_orig)

    # compare_orig_mask_gt_pred(im_orig, im_gt, im_orig,"comparing originals")



from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np


mpl.rcParams["figure.dpi"] = 300


import tifffile
from scipy.ndimage import label


def bin_2d(im, bin_size, stride=1, pad_value=0, debug=False):
    """
    Bins a 2d array by a square kernel length (bin_size*2 + 1).

    Parameters
    ----------
    im : ndarray
        3d ndarray containing FLIM data (x,y,t).
    bin_size : int
        number of pixels to look before and after center pixel.
        kernel = bin_size + 1 + bin_size
    stride : int, optional
        Number of pixels to move when raster scanning the image. The default is 1.
    pad_value : int, optional
        value to pad the image with. The default is 0.
    debug : bool, optional
        Show debugging output. The default is False.

    Returns
    -------
    im_binned : ndarray
        binned 3d array of decays

    """

    if stride != 1:
        print("strides larger than 1 not yet implemented")

    kernel_length = bin_size * 2 + 1  # bin left center pixel and right sides

    rows_before = bin_size
    rows_after = bin_size
    cols_before = bin_size
    cols_after = bin_size

    n_rows, n_cols = im.shape

    padded_im = np.pad(
        im,  # this takes in width(cols) and height(rows)
        ((rows_before, rows_after), (cols_before, cols_after)),
        "constant",
        constant_values=pad_value,
    )

    im_binned = np.zeros(im.shape)

    # iterate through original image dimensions
    for row_idx in np.arange(n_rows):
        for col_idx in np.arange(n_cols):

            # starts at 0(row_idx) : kernel_length + 0
            # then 1 (row_idx) : kernel_length + 1 ...etc
            # up to length n-bin_size
            # get window on padded image
            conv_decay = padded_im[
                row_idx : kernel_length + row_idx, col_idx : kernel_length + col_idx
            ]
            # fill in original shape image
            im_binned[row_idx, col_idx] = conv_decay.sum(axis=(0, 1))

    if debug:
        plt.title(f"kernel: {kernel_length}x{kernel_length}")
        plt.imshow(im_binned)
        plt.show()

    return im_binned


if __name__ == "__main__":

    ### bin_2d TEST
    HERE = Path(__file__).resolve().parent

    # load SDT
    im = np.random.rand(512, 512)
    # plt.title("original")
    # plt.imshow(im)
    # plt.show()

    # run with debug on
    bin_size = 2
    im_binned = bin_2d(im, bin_size=bin_size, debug=False)
    
    from cell_analysis_tools.visualization import compare_images
    
    compare_images("original", im, f"binned {bin_size*2_1}x{bin_size*2_1}", im_binned)



import codecs
import re
from pathlib import Path

import numpy
import numpy as np
import tifffile
from matplotlib import pyplot as plt
from read_roi import read_roi_zip
from skimage.draw import polygon2mask


def read_asc(path):
    """
    Reads in an asc file into a numpy ndarray

    Parameters
    ----------
    path : pathlib.Path
        path to the file.

    Returns
    -------
    array : np.ndarray
        Numpy array holding the image data.

    """
    with codecs.open(path, encoding="utf-8-sig") as file:
        # for each line for each value, convert to float
        array = np.array([[float(x) for x in line.split()] for line in file])

    return array


if __name__ == "__main__":

    ### test loading ASC files
    test_files = Path("./test_files")
    test_files.exists()
    list_path_test_files = list(test_files.glob("*"))
    list_test_files_str = [str(p) for p in list_path_test_files]

    list_asc_files_str = list(filter(re.compile(r".*\.asc").match, list_test_files_str))
    list_asc_files = [Path(p) for p in list_asc_files_str]

    for path_file in list_asc_files:
        image = read_asc(path_file)
        plt.title(path_file.name)
        plt.imshow(image)
        plt.show()



# Dependencies
import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def phasor_to_rectangular_lifetimes_array(angles, magnitudes):
    """ Takes in two arrays of angles and magnitudes
    and converts them to array of g and s points
    
    Parameters
    ----------
        f : int 
            laser angular frequency
        angles : float 
            array of angles
        mangitudes : float
            array of magnitudes
    Returns
    -------
        RectangularLifetimesArray : Phasor collection object
            object with g and s array of lifetimes
                g - array of g-coordinates(x-coordinates)
                s - array of s-coordinates (y-coordinates)
    """
    RectangularLifetimesArray = coll.namedtuple("RectangularLifetimesArray", "g s")

    # calculate g and s
    g = magnitudes * np.cos(angles)
    s = magnitudes * np.sin(angles)

    return RectangularLifetimesArray(g=g, s=s)



from .image_viewers import compare_images, compare_orig_mask_gt_pred, image_show
from .umap_tsne_pca import compute_pca, compute_tsne, compute_umap

__all__ = [
    "image_show",
    "compare_images",
    "compare_orig_mask_gt_pred",
    "compute_pca",
    "compute_tsne",
    "compute_umap",
]



import numpy as np
from numpy.ma import masked_array
from skimage.measure import label
from skimage.measure import regionprops as _regionprops
from skimage.measure import regionprops_table as _regionprops_table
from skimage.morphology import label

from .fractal_dimension.fractal_dim_gray import fractal_dimension_gray
from .intensity_sum import intensity_sum
from .roi_distance import radius_max, radius_mean, radius_median


def regionprops(label_image, intensity_image=None):
    """ Extended regionprops function adding our own props    
    
    To see a complete docstring for this function see regionprops skimage.

    Notes
    -----
    Additional properties that can be accessed as attributes or keys:

    **radius_max** : float
        The maximum distance of any pixel in the segmented region to the closest background pixel.
    **radius_mean** : float
        Mean value the distances of all pixels in the segmented region to their closest background pixel.
    **radius_median** : float
        Median value the distances of all pixels in the segmented region to their closest background pixel.
    **intensity_sum**
        The sum of the pixel intensities within the segmented region.
    **fractal_dimension** : float
        Fractal dimension, differential box counting method implementation

    """
    return _regionprops(
        label_image,
        intensity_image=intensity_image,
        extra_properties=(
            radius_mean,
            radius_median,
            radius_max,
            intensity_sum,
            fractal_dimension_gray,
        ),
    )


def regionprops_table(label_image, intensity_image, properties=None):

    return _regionprops_table(
        label_image,
        intensity_image=intensity_image,
        properties=properties,
        extra_properties=(
            radius_mean,
            radius_median,
            radius_max,
            intensity_sum,
            fractal_dimension_gray,
        ),
    )


if __name__ == "__main__":

    import matplotlib as mpl
    import matplotlib.pylab as plt
    from skimage.draw import ellipse

    # import numpy as np

    mpl.rcParams["figure.dpi"] = 300

    idx_rows, idx_cols = ellipse(20, 20, 5, 7)
    shape_ellipse = np.zeros((40, 40))
    shape_ellipse[idx_rows, idx_cols] = 1

    import numpy as np

    rng = np.random.default_rng(seed=0)

    intensity = rng.random(shape_ellipse.shape) * shape_ellipse
    intensity = intensity * 255

    plt.imshow(shape_ellipse)
    plt.show()
    shape_ellipse = shape_ellipse.astype(int)

    labels = label(shape_ellipse)
    props = regionprops_table(
        labels,
        intensity,
        properties=[
            "area",
            "major_axis_length",
            "minor_axis_length",
            "eccentricity",
            "orientation",
            "solidity",
            "extent",
            "perimeter",
            "radius_max",
            "radius_mean",
            "radius_median",
            "fractal_dimension_gray",
        ],
    )



from .bin_2d import bin_2d
# from .bin_3d import bin_3d
from .fft_image_filter import remove_horizontal_vertical_edges
from .fill_and_label_rois import fill_and_label_rois
from .four_color_theorem.four_color_theorem_to_unique_values import four_color_to_unique
from .four_color_theorem.four_colors import four_color_theorem
from .kmeans_threshold import kmeans_threshold
from .normalize import normalize
from .rgb2gray import rgb2gray
from .rgb2labels import rgb2labels
from .sum_pool_3d import sum_pool_3d
from .remove_small_areas_fill_regions import remove_small_areas_fill_regions

__all__ = [
    "bin_2d",
    # "bin_3d",
    "sum_pool_3d",
    "normalize",
    "kmeans_threshold",
    "rgb2labels",
    "rgb2gray",
    "remove_horizontal_vertical_edges",
    "fill_and_label_rois",
    "four_color_theorem",
    "four_color_to_unique",
    "remove_small_areas_fill_regions"
]



import scipy.ndimage as ndi


def fill_and_label_rois(curr_nuclei):
    """
    Fills and labels ROI outlines using unique ints for each region.

    Parameters
    ---------- 
    param curr_nucelei : ndarray
        Current image to process

    Returns
    -------
    output : ndarray 
        ROIs filled and labeled with unique int representations

    """

    return ndi.label(ndi.binary_fill_holes(curr_nuclei))[0]

    
if __name__ == "__main__":
    
    from pathlib import Path
    
    
    HERE = Path(__file__).absolute().resolve()
    print(HERE)
    path_mask = Path(r"../../examples/example_data/redox_ratio/HPDE_2DG_10n_mask_cells.tiff")
    
    
    


import numpy as np
from .helper import _validate_array_and_make_bool


def jaccard(mask_gt, mask_pred):
    """
    Calculates the IoU/jaccard index for a pair of masks. 
    
    Logic is implemented to require images of the same size. 

    Parameters
    ----------
    mask_pred: np.ndarray
        Ground truth numpy ndarray image.
    mask_gt: np.ndarray
        Current image numpy ndarray.

    Returns
    -------
    jaccard_index: float
        Calculated Jaccard index/distance.
    """

    if mask_gt.shape != mask_pred.shape:
        raise ValueError(
            f"Shape mismatch: the shape of the ground truth mask {mask_gt.shape} does not match shape of predicted mask {mask_pred.shape}"
        )

    if not mask_gt.any() and not mask_pred.any():
        raise ValueError(
            "Ground truth mask and predicted mask cannot both be entirely zeros"
        )

    mask_gt = _validate_array_and_make_bool(mask_gt)
    mask_pred = _validate_array_and_make_bool(mask_pred)

    return (
        np.logical_and(mask_gt, mask_pred).sum()
        / np.logical_or(mask_gt, mask_pred).sum()
    )



from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np

from cell_analysis_tools.io import load_sdt_file

mpl.rcParams["figure.dpi"] = 300


def sum_pool_3d(im, bin_size, stride=1, pad_value=0, debug=False):
    """
    Sums the pixel intensity by a given kernel, reducing output image dimensions.

    Parameters
    ----------
    im : ndarray
        3d array containig FLIM data (x,y,t).
    bin_size : int
        number of pixels to select before and after, bin of 3 is a 7x7 kernel
    stride : int, optional
        number of pixels to advance when doing raster calculations. The default is 1.
    pad_value : int, optional
        Value to pad the edges with if kernel needs it. The default is 0.
    debug : bool , optional
        Show debugging output. The default is False.

    Returns
    -------
    im_sum_pool : ndarray
        3d ndarray that holds a summed version of the image resized image.

    """
    if stride != 1:
        print("strides larger than 1 not yet implemented")

    kernel_length = bin_size * 2 + 1  # 7 kernel length
    # convolve with surrounding pixels

    ## Make sure image is right size or pad
    im_rows, im_cols, im_timebins = im.shape  # collapse and get 2d shape

    # COMPUTE PADDING PIXELS
    remainder_n_row_pixels = (im_rows / stride) % kernel_length
    remainder_n_col_pixels = (im_cols / stride) % kernel_length

    if remainder_n_row_pixels != 0:
        need_n_row_pixels = kernel_length - remainder_n_row_pixels
    else:
        need_n_row_pixels = 0

    if remainder_n_col_pixels != 0:
        need_n_col_pixels = kernel_length - remainder_n_col_pixels
    else:
        need_n_col_pixels = 0

    # pad image at the end of rows and cols

    rows_before = 0
    rows_after = int(need_n_row_pixels)
    cols_before = 0
    cols_after = int(need_n_col_pixels)

    padded_sdt = np.pad(
        im,  # this takes in width(cols) and height(rows)
        (
            (rows_before, rows_after),
            (cols_before, cols_after),
            (0, 0),  # dont pad 3rd dim
        ),
        "constant",
        constant_values=pad_value,
    )

    new_im_rows, new_im_cols, _ = padded_sdt.shape

    # new image placeholder
    im_sum_pool = np.zeros(
        (
            int(new_im_rows / kernel_length),
            int(new_im_cols / kernel_length),
            im_timebins,
        )
    )
    new_n_rows, new_n_cols, _new_n_timebins = im_sum_pool.shape

    # bin image
    for row in np.arange(new_n_rows):
        for col in np.arange(new_n_cols):
            pass
            sdt_block = padded_sdt[
                row * kernel_length : row * kernel_length + kernel_length,
                col * kernel_length : col * kernel_length + kernel_length,
                :,
            ]
            im_sum_pool[row, col, :] = np.sum(sdt_block, axis=(0, 1))

    if debug:
        plt.title(f"kernel_shape: {kernel_length}x{kernel_length}")
        plt.imshow(im_sum_pool.sum(axis=2))
        plt.show()

    return im_sum_pool


if __name__ == "__main__":

    ### sum_pool TEST
    HERE = Path(__file__).resolve().parent
    data = HERE.parent / r"example_data\t_cell\Tcells-001.sdt".replace("\\", "/")

    # load SDT
    im = load_sdt_file(data)
    for ch in np.arange(im.shape[0]):
        plt.title(f"channel: {ch}")
        plt.imshow(im[ch, ...].sum(axis=2))
        plt.show()

    # select channel with data
    im = im[1, ...]

    # run with debug on
    sum_pool_3d(im, bin_size=2, debug=True)



from time import time

import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np
from numpy import ones
from numpy.fft import fft2, fftshift, ifft, ifft2, ifftshift
from PIL import Image
from skimage.morphology import disk

mpl.rcParams["figure.dpi"] = 300


def remove_horizontal_vertical_edges(im, disk_size=20, debug=False):
    """ This function removes horizontal or vertical edges in the image. Designed for removing grid patterns in
    PDMS scaffolds during time lapse imaging. 

    Parameters
    ----------
    im : ndarray
        intensit image with grid pattern
    disk_size : int
        size of filter used to filter low frequency components of 2d image
    debug : bool, optional
        output intermediate images , by default False

    Returns
    -------
    np.ndarray
        filtered image with grid pattern removed
        
    
    .. image:: ./resources/image_processing-remove_horizontal_vertical_edges.png
        :width: 400
        :alt: Image of grid removed through fft filter
    """

    tic = time()

    imsize = min(im.shape)
    im = im[:imsize, :imsize]  # make square
    # % Fourier transform of the image (complex array)
    # fftim = fftshift(fft2(fftshift(im)));
    fftim = fftshift(fft2(fftshift(im)))

    # % create a mask to filter out certain spatial frequencies
    # filter = circ(imsize,20); % pass low freqs near the origin
    selem_disk = disk(disk_size)
    pad_left = (im.shape[0] - selem_disk.shape[0]) // 2
    pad_right = int(np.ceil((im.shape[0] - selem_disk.shape[0]) / 2))

    filt1 = np.pad(
        selem_disk, (pad_left, pad_right), mode="constant", constant_values=0
    )

    ##
    # spokewidth = 5; % width of spokes, bigger catches more angles. Could do a pinwheel or other more sophisticated shapes.
    # filter2 = ones(size(im)); filter2(imsize/2-spokewidth:imsize/2+spokewidth,:)=0;filter2(:,imsize/2-spokewidth:imsize/2+spokewidth)=0;
    # filter = (filter|filter2);

    spokewidth = 5  # ; % width of spokes, bigger catches more angles. Could do a pinwheel or other more sophisticated shapes.
    filt2 = ones(im.shape)
    filt2[imsize // 2 - spokewidth : imsize // 2 + spokewidth, :] = 0
    filt2[:, imsize // 2 - spokewidth : imsize // 2 + spokewidth] = 0
    filt2 = filt2.astype(np.uint8)
    filt = [filt1 | filt2]

    #% apply the filter in Fourier space and FFT back to image space
    # filteredim = abs(fftshift(ifft2(ifftshift(fftim.*filter))));
    filteredim = abs(fftshift(ifft2(ifftshift(fftim * filt))))

    if debug:
        toc = time()
        print(f"elapsed time: {toc-tic}")

        ## display
        fig, ax = plt.subplots(1, 2)
        ax[0].imshow(im, cmap="gray")
        ax[0].set_title("original")
        ax[1].imshow(filteredim.squeeze(), cmap="gray")
        ax[1].set_title("Fourier filtered")

    return filteredim


if __name__ == "__main__":

    im = Image.open(r"scaffold.png".replace("\\", "/"))
    im = np.asarray(im)

    # plt.imshow(im)
    im = im.sum(axis=2)
    im = im[:, :1018]
    print(im.shape)
    im_filtered = remove_horizontal_vertical_edges(im, debug=True)



import numpy as np
from scipy import ndimage


# https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.distance_transform_edt.html#scipy.ndimage.distance_transform_edt
def radius_max(roi, return_index=False):
    """

    The maximum distance of any pixel in the segmented region to the closest background pixel.


    Parameters
    ----------
    roi : ndarray
        binary roi mask.
    return_index : bool, optional
        return index of pixel where max radius was found. The default is False.
        
    Returns
    -------
    float
        Max eucledian distance of roi to background.
        
    Note
    ----
    If there are multiple, it returns the first instance (see np.argmax)

    References
    ----------
    """
    distance, indices = ndimage.distance_transform_edt(roi, return_indices=True)
    idx_max_value = np.unravel_index(distance.argmax(), roi.shape)

    if return_index:
        return distance[idx_max_value], idx_max_value

    return distance[idx_max_value]


def radius_mean(roi):
    """
    Mean value the distances of all pixels in the segmented region to their closest background pixel.

    Parameters
    ----------
    roi : ndarray
        binary roi mask.

    Returns
    -------
    float
        mean distance of all the pixel distances to background

    """

    return np.mean(ndimage.distance_transform_edt(roi))


def radius_median(roi):
    """
    Median value the distances of all pixels in the segmented region to their closest background pixel.

    Parameters
    ----------
    roi : ndarray
        binary roi mask.

    Returns
    -------
    float
        median distance of all the pixels distances to background

    """
    return np.median(ndimage.distance_transform_edt(roi))


if __name__ == "__main__":

    import matplotlib as mpl
    import matplotlib.pylab as plt

    mpl.rcParams["figure.dpi"] = 300
    import numpy as np
    from skimage.morphology import disk

    roi = disk(20)
    plt.imshow(roi)
    plt.show

    # MAX RADIUS
    # https://stackoverflow.com/questions/9482550/argmax-of-numpy-array-returning-non-flat-indices
    distance, indices = ndimage.distance_transform_edt(roi, return_indices=True)
    idx_max_value = np.unravel_index(distance.argmax(), distance.shape)
    distance[idx_max_value]

    print(f"max radius: {radius_max(roi)}")
    # MEAN RADIUS
    print(f"mean radius: {radius_mean(roi)}")

    # MEDIAN RADIUS
    print(f"median radius: {radius_median(roi)}")



# Dependencies
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def draw_universal_semicircle(
    laser_angular_frequency, title="Phasor Plot", suptitle="", debug=False
    ):
    """
    Draws the universal semicircle with the given frequency and returns the plt
    figure to draw over.

    Parameters
    ----------
        figure : matplotlib figure object
            figure object to draw semicircle over
        laser_angular_frequency : int
            rep rate or laser angular frequency, affects position of labeled points

    Returns
    -------
        plt.figure object to plot your points over this figure
        
    .. image:: ./resources/flim_phasor_plot_80.png
        :width: 400
        :alt: Image of universal semicircle at 80Mhz''
        
    .. image:: ./resources/flim_phasor_plot_100.png
        :width: 400
        :alt: Image of universal semicircle at 100Mhz''
    
    
    """
    
    figure = plt.figure()
    plt.title(title)
    """ get universal semicircle values for this rep rate """
    x_circle, y_circle, g, s, lifetime_labels = universal_semicircle_series(
        laser_angular_frequency
    )

    # Labels and axis of phasor plot
    # figure = plt.figure()
    figure.suptitle(suptitle)
    plt.xlabel("g", fontsize=20)
    plt.ylabel("s", fontsize=20)

    # add circle and lifetime estimates
    plt.plot(x_circle, y_circle, "-", color="teal")
    plt.plot(g, s, ".", color="magenta")

    if debug:
        print("type: ", type(lifetime_labels), " labels: ", lifetime_labels)

    """ ADD LIFETIME LABELS """
    for i, txt in enumerate(lifetime_labels):
        # self.ax.annotate(txt, (g[i], s[i]))
        plt.annotate(txt, (g[i], s[i]))

    plt.axis('equal')
    # plt.savefig(f"flim_phasor_plot_{int(frequency/1e6)}.png",bbox_inches='tight')
    # plt.show()


def universal_semicircle_series(frequency):
    """
    With the given frequency, this function will return x and y points for plotting the
    universal semicircle, as well as the corresponding g and s values for
    common lifetimes between 0.5ns to 10ns

    Parameters
    ----------
        frequency {float} : angular laser repetition frequency
    
    Returns
    -------

        x_circle : float
            x coordinates for universal semicircle
        y_circle : float
            y coordinates for universal semicircle
        g : float
            x coordinates for labeled lifetimes
        s : float
            y coordinates for labeled lifetimes
        lifetime_labels : list
            labels to be applied to the (g,s) coordinates
    """
    x_coord = np.linspace(0, 1, num=1000)
    y_circle = np.sqrt(x_coord - x_coord ** 2)
    x_circle = x_coord ** 2 + y_circle ** 2

    omega = 2.0 * np.pi * frequency  # modulation frequency
    tau = np.asarray(
        [0.5e-9, 1e-9, 2e-9, 3e-9, 4e-9, 5e-9, 6e-9, 7e-9, 8e-9, 9e-9, 10e-9]
    )  # lifetimes in ns
    g = 1 / (1 + np.square(omega) * np.square(tau))
    s = (omega * tau) / (1 + np.square(omega) * np.square(tau))
    lifetime_labels = [
        "0.5ns",
        "1ns",
        "2ns",
        "3ns",
        "4ns",
        "5ns",
        "6ns",
        "7ns",
        "8ns",
        "9ns",
        "10ns",
    ]
    
    # lifetime_labels2 = [f"{t*1e9:.1f}ns" for t in tau]
    return x_circle, y_circle, g, s, lifetime_labels


if __name__ == "__main__":
    
    frequency = 100e6
    draw_universal_semicircle( 
                              laser_angular_frequency=frequency,
                              title=f"Phasor Plot at {frequency/1e6:.0f} Mhz")



from pathlib import Path

import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np

mpl.rcParams["figure.dpi"] = 300


import tifffile
from scipy.ndimage import label


def rgb2labels(im, debug=False):
    """
    Converts a 3ch rgb mask image into a 1ch uint16 mask image.   

    Parameters
    ----------
    im : ndarray
        input 3ch RGB image of rois.
    debug : bool, optional
        Enable display of intermediates. The default is False.

    Returns
    -------
    mask : ndarray
        labels mask from RGB image.

    """
    if debug:
        plt.title("original_image")
        plt.imshow(im)
        plt.show()

    ##### convert RGB to intensity with unique values
    im[..., 1] = im[..., 1] * 2
    im[..., 2] = im[..., 2] * 3
    flat = np.sum(im, axis=2)

    if debug:
        plt.title("flattened image")
        plt.imshow(flat)
        plt.show()

    unique_values = np.unique(flat)
    unique_values = unique_values[1:]  # remove bg
    if debug:
        print(f"unique values without bg: {unique_values}")

    # output mask
    mask = np.zeros(im.shape[:2])

    # incrementing index
    idx = 1
    for value in unique_values:  # iterate through values
        im_rois = (flat == value).astype(int)
        labeled_mask, n_rois = label(im_rois)
        # iterate through labels
        for roi_label in np.unique(labeled_mask)[1:]:  # exclude gb mask
            roi = (labeled_mask == roi_label).astype(int) * idx
            mask = mask + roi
            idx += 1
    if debug:
        plt.title("intermediate mask")
        plt.imshow(mask)
        plt.show()

    ###### reorder index to increasing values

    iter_mask = mask.copy().astype(int)
    n_rows, n_cols = iter_mask.shape

    label_id = 1  # starting label
    for row in np.arange(n_rows):
        print(f"processing row: {row}")
        for col in np.arange(n_cols):
            if iter_mask[row, col] != 0:  # mask found
                # print(f"value: {mask[row,col]}")
                roi_value = mask[row, col]
                mask[iter_mask == roi_value] = label_id  # replace mask value
                iter_mask[iter_mask == roi_value] = 0  # clear out mask in iter_mask
                label_id += 1  # increment index
    if debug:
        plt.title("incrementing order of labels")
        plt.imshow(mask)
        plt.show()

    return mask.astype(np.uint16)


if __name__ == "__main__":

    #### rgb2labels test

    # https://stackoverflow.com/questions/24780697/numpy-unique-list-of-colors-in-the-image
    # pixels_1d_array = im.reshape((-1,im.shape[2]))
    # unique_colors = np.unique(pixels_1d_array, axis=0)

    image_dir = Path("rgb2gray/sample_data")

    # path_im = image_dir / "N_photons_day22_feat2_Object Predictions_cyto_overlay.tiff"
    path_im = image_dir / "N_photons_day22_feat2_Object Predictions_cyto.tiff"

    debug = True
    im = tifffile.imread(path_im)
    mask = rgb2labels(im, debug=True)

    unique_colors = np.unique(mask, axis=0)



#!/usr/bin/env python

from pathlib import Path

from setuptools import find_packages, setup

THIS_DIR = Path(__file__).parent

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="cell_analysis_tools",
    version="0.0.14",
    author="Emmanuel Contreras Guzman",
    author_email="econtreras@wisc.edu",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
    ],
    packages=find_packages(exclude=["tests", "dev"]),
    # install_requires=parse_requirements("requirements.txt").read_text().splitlines(),
    install_requires=[
    'read-roi'
    'umap-learn'
    'scikit-image'
    'scikit-learn'
    'matplotlib'
    ]
)



import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def ideal_sample_phasor(f, lifetime):
    """
    Generates a phasor for a specific lifetime at a given frequency.

    Parameters
    ----------
        f : int
            laser rep rate
        lifetime : float
            lifetime of desired single exponential sample

    Returns
    -------
        angle : float
            angle of ideal sample
        magnitude : float
            magnitude of ideal sample
            
    .. code-block:: python
        
        >>print("Input is 80Mhz phasor at @ 2ns lifetime")
        >>angle, magnitude = ideal_sample_phasor(f=80e6, lifetime=2e-9)
        >>print(f"{angle=:.3f} rad | {magnitude=:.3f}")
        angle=0.788 rad | magnitude=0.705
        
    
    
    """

    Phasor = coll.namedtuple("Phasor", "angle magnitude")

    # lifetime = lifetime * 1e-12  # lifetime in ns
    w = 2 * np.pi * f
    ### simulated point values
    ideal_g = 1 / (1 + (w ** 2 * lifetime ** 2))
    ideal_s = (w * lifetime) / (1 + (w * lifetime) ** 2)

    # angle = np.arctan(ideal_s/ideal_g) # radians
    angle = np.pi - np.arctan2(ideal_s, -ideal_g)  # in radians
    magnitude = np.sqrt(ideal_g ** 2 + ideal_s ** 2)

    print("Input lifetime: ", (1 / w * ideal_s / ideal_g))
    return Phasor(angle=angle, magnitude=magnitude)

if __name__ == "__main__":
    
    from cell_analysis_tools.flim import ideal_sample_phasor
    
    angle, magnitude = ideal_sample_phasor(80e6, 2e-9)
    
    print(f"{angle=:.3f} rad | {magnitude=:.3f}")
    
    from cell_analysis_tools.flim import draw_universal_semicircle
    
    fig = draw_universal_semicircle(laser_angular_frequency=80e6)
    
    plt.scatter(0.4, 0.2, c='k')
    plt.show()
    
    



from .load_image import load_image
from .load_sdt import load_sdt_data, load_sdt_file
from .read_asc import read_asc

__all__ = [
    "load_sdt_data",
    "load_sdt_file",
    "read_asc",
    "load_image",
]



# Dependencies
import collections as coll

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import pylab
import tifffile
from scipy.signal import convolve

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc


def lifetime_image_to_rectangular_points(f, image):
    """
    This function takes all the pixels/histograms
    of an image and plots them in the phasor plot

    Parameters
    ----------
    f : frequency 
        laser angular frequency
    image : ndarray
        3d array representation of image, where the axis are [t,x,y]

    Returns
    -------

    ImagePhasorPoints : ndarray 
        object holding array of points
            points_g - 2d array of g coordinates
            points_s - 2d array of s coordinates
    """

    # use the code below created a 3d "image" of histograms with (np.cos(w * timebins)) pre computed
    # then do matrix multiplication to calculate the g and s for every pixel and then plot

    ImagePhasorPoints = coll.namedtuple("ImagePhasorPoints", "points_g points_s")
    # print(image.shape)
    width, height, num_timebins = image.shape
    laser_period = 1 / f
    timebins = np.linspace(0, laser_period, num_timebins, endpoint=False)
    w = 2 * np.pi * f

    """ create 3d matrix of """
    # https://stackoverflow.com/questions/24148322/python-3d-array-times-1d-vector
    #    ones_array = np.ones((num_timebins))
    pre_comp_cos = np.cos(w * timebins)
    pre_comp_sin = np.sin(w * timebins)
    # newaxis == None, helps index/align the arrays for multiplication
    # cos_array = ones_array * pre_comp_cos[:,np.newaxis,np.newaxis]
    # sin_array = ones_array * pre_comp_sin[:,None,None]

    # image = np.ones((4,4,4)) * np.linspace(1,4,4)[:,None,None]
    # te = np.array([4,3,2,1])
    # temp = image * te[:,np.newaxis,np.newaxis]

    """image arrangement """
    axis_x, axis_y, axis_timebins = (0, 1, 2)

    #     point_g = np.sum(counts * np.cos(w * timebins)) / np.sum(counts)
    with np.errstate(divide="ignore", invalid="ignore"):
        # numerator = np.sum(image * pre_comp_cos[np.newaxis, np.newaxis,:], axis = axis_timebins)
        numerator = image
        denominator = np.sum(image, axis=axis_timebins)

        temp = np.true_divide(numerator, denominator[:, :, np.newaxis])
        temp = np.nan_to_num(temp)  # check 0/0, NaN=0
        points_g = np.sum(
            temp * pre_comp_cos[np.newaxis, np.newaxis, :], axis=axis_timebins
        )

        # points_g = np.true_divide(numerator, denominator)
        # print('division', points_g)
        # sometimes will divide by 0/0, 1/0, -1/0 check all cases
        # points_g = np.nan_to_num(points_g) # check 0/0, NaN=0
        # points_g[~ np.isfinite(points_g)] = 0 #inf and -inf = 0

    with np.errstate(divide="ignore", invalid="ignore"):
        # numerator = np.sum(image * pre_comp_sin[np.newaxis, np.newaxis, :], axis=axis_timebins)

        # temp = np.true_divide(numerator, denominator[:,:, np.newaxis])
        # temp = np.nan_to_num(temp)  # check 0/0, NaN=0
        points_s = np.sum(
            temp * pre_comp_sin[np.newaxis, np.newaxis, :], axis=axis_timebins
        )

        # denominator = np.sum(image, axis=axis_timebins)
        # points_s = np.true_divide(numerator, denominator)
        # sometimes will divide by 0/0, 1/0, -1/0 check all cases
        # points_s = np.nan_to_num(points_s)
        # points_s[~ np.isfinite(points_s)] = 0

    return ImagePhasorPoints(points_g=points_g, points_s=points_s)


if __name__ == "__main__":
    
    
    
    from cell_analysis_tools.io import load_sdt_file
    from cell_analysis_tools.flim import (phasor_calibration, 
                                          rectangular_to_phasor_lifetimes_array)
    
    from cell_analysis_tools.flim import draw_universal_semicircle, phasor_calibration
    
    from copy import copy
    # # load irf
    irf_decay = np.loadtxt("irf.csv")
    timebins = irf_decay[:,0]
    irf = irf_decay[:,1]
    plt.plot(timebins, irf)
    
    # load image 
    sdt = load_sdt_file("./resources/test_image.sdt")
    im_nadh = sdt[1,...]
    plt.imshow(im_nadh.sum(axis=2))
    plt.show()
    
       # compute g and s 
    g,s = lifetime_image_to_rectangular_points(80e6, im_nadh)
    
    # convert to phasors
    im_phasor = rectangular_to_phasor_lifetimes_array(g=g,s=s)
    
    # calibrate
    calibration = phasor_calibration(f=80e6, lifetime=2e-9, timebins=timebins, counts=im_nadh)
    

    draw_universal_semicircle(80e6)    
    im_phasor_calibrated_angles = im_phasor.angles - calibration.angle
    im_phasor_calibrated_magnitudes = im_phasor.magnitudes * calibration.scaling_factor
    
    
    new_g, new_s = phasor_to_rectangular_lifetimes_array ()
    

    plt.scatter(g,s, s=1)
    plt.show()    
    
    


# MATLAB
# function [M, Phi, G, S] = phasor_calculator(f, time, decays, IRF)
# w = 2*pi*f;
# G_decay = (decays*cos(w*time))./sum(decays,2);
# S_decay = (decays*sin(w*time))./sum(decays,2);
# G_IRF = sum(IRF.*cos(w*time))/sum(IRF(:));
# S_IRF = sum(IRF.*sin(w*time))/sum(IRF(:));
# GS = [G_IRF, S_IRF; -S_IRF, G_IRF]*[G_decay'; S_decay']/(G_IRF^2 + S_IRF^2);
# G = transpose(GS(1,:));
# S = transpose(GS(2,:));
# Phi = pi-atan2(S,-G);
# M = sqrt(G.^2 + S.^2);
# end


import matplotlib.pylab as plt
import matplotlib as mpl
mpl.rcParams['figure.dpi']=300
from cell_analysis_tools.io import load_sdt_file

import numpy as np
from cell_analysis_tools.flim import draw_universal_semicircle


def phasor_calculator(f, time, decays, IRF):
    """
    Given an array of decay(s) the rectangular g, s and phasors angle and magnitude 
    will be computed and returned.

    Parameters
    ----------
    f : float
        Laser repetition rate.
    time : np.ndarray
        array of timebins.
    decays : np.ndarray
        Single decay or array of decays to compute phasor points for.
        Single decay should have the shape (x,),(x,t) and shape (x,y,t) for an array
    IRF : np.ndarray
        1D array capturing irf decay.

    Returns
    -------
    m : float
        magnidue of phasor.
    phi : float
        angle of phasor.
    g : float
        g coordinate (x-axis).
    s : float
        s coordinate (y-axis).


    Note
    ----
        * You cannot compare two images directly due to them having different decay shift values between images, affeting g,s,m and phi locations 
        * For proper lifetime values, background subtraction is needed by taking ~ the last 1 or 1/2 ns timebins of decay)
    
        
    .. image:: ./resources/flim_phasor_calculation.png
        :width: 600
        :alt: Image of a lifetime image as a phasor plot
        
        
    """
    
    w = 2*np.pi*f
    
    # row=pixels, cols=photons
    decays = np.reshape(decays,(-1,256))
    
    
    dim_time = len(decays.shape) -1
    G_decay = np.dot(decays, np.cos(w*time))/np.sum(decays, axis=dim_time)
    S_decay = np.dot(decays, np.sin(w*time))/np.sum(decays, axis=dim_time)
    G_IRF = np.dot(IRF,np.cos(w*time))/np.sum(IRF)
    S_IRF = np.dot(IRF, np.sin(w*time))/np.sum(IRF)
    GS = np.dot(np.array([[G_IRF, S_IRF], [-S_IRF, G_IRF]]), np.array([G_decay, S_decay]))/(G_IRF**2 + S_IRF**2)
    
    g = GS[0,:]
    s = GS[1,:]
    
    phi = np.pi-np.arctan2(s,-g)
    m = np.sqrt(g**2 + s**2)
    
    return m, phi, g, s


# for image pixels by time

if __name__ == "__main__":
    # # load irf
    irf_decay = np.loadtxt("irf.csv")   
    timebins = irf_decay[:,0]
    irf = irf_decay[:,1]
    plt.plot(timebins,irf)
    plt.title(f"IRF")
    plt.show()
    
    sdt = load_sdt_file("./bigger_beads_2.1ns.sdt")
    im_nadh = sdt.squeeze()
    decay = im_nadh.sum(axis=(0,1))
    plt.title("Beads Decay")
    plt.plot(decay)
    plt.show()
    
    frequency = 0.08
    draw_universal_semicircle(laser_angular_frequency=frequency*10**9,
                              title=f"bead image as 1 decay",
                              suptitle="Phasor Plot")
    m, phi, g, s = phasor_calculator(f=frequency, time=timebins, decays=decay, IRF=irf)
    plt.scatter(g,s, label="single point", s=3)
    plt.show()
    
    
    # threshold image
    mask = im_nadh.sum(axis=2) > 1000
    im_phasor = im_nadh * mask[:,:,np.newaxis]
    
    
    # compute g,s and plot
    m, phi, g, s = phasor_calculator(f=frequency, time=timebins, decays=im_phasor, IRF=irf)
    
    draw_universal_semicircle(laser_angular_frequency=frequency*10**9,
                              suptitle="Phasor Plot",
                              title="bead image")
    plt.scatter(g,s, s=1)
    plt.show()
    



#%%
from pprint import pprint

import matplotlib as mpl
import matplotlib.pylab as plt
import numpy as np
import numpy.ma as ma
import tifffile
from skimage.measure import regionprops
from skimage.morphology import label

from cell_analysis_tools.image_processing import normalize
from cell_analysis_tools.io import read_asc

mpl.rcParams["figure.dpi"] = 300
import math


#%%
def regionprops_omi(
    image_id: str,  # base_name of image
    label_image: np.ndarray,
    im_nadh_intensity: np.ndarray = None,
    im_nadh_a1: np.ndarray = None,
    im_nadh_a2: np.ndarray = None,
    im_nadh_t1: np.ndarray = None,
    im_nadh_t2: np.ndarray = None,
    im_fad_intensity: np.ndarray = None,
    im_fad_a1: np.ndarray = None,
    im_fad_a2: np.ndarray = None,
    im_fad_t1: np.ndarray = None,
    im_fad_t2: np.ndarray = None,
    im_nadh_chi: np.ndarray = None,
    im_fad_chi: np.ndarray = None,
    other_props: list = None,
) -> dict:
    #%%
    """
    Takes in labels image as well as nadh and fad images from SPCImage to return
    mean and stdev of each parameter per roi. 

    Parameters
    ----------
        label_image : ndarray
            labeled mask image.
        im_nadh_intensity : ndarray
            nadh intensity image.
        im_nadh_a1 : ndarray
            nadh alpha1 image.
        im_nadh_a2 : ndarray
            nadh alpha2 image .
        im_nadh_t1 : ndarray
            nadh tau 1 lifetime, short .
        im_nadh_t2 : ndarray
            nadh tau 2 lifetime, long.
        im_fad_intensity : ndarray
            nadh intensity image. 
        im_fad_a1 : ndarray
            fad alpha 1 image. 
        im_fad_a2 : ndarray
            fad alpha 2 image. 
        im_fad_t1 : ndarray
            fad tau 1 lifetime, long. 
        im_fad_t2 : ndarray
            fad tau 2 lifetime, short.
        other_props : list
            string list of additional parameters to compute on the binary mask
            see skimage regionprops for list of attributes 
    
    .. note::
        See `https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.regionprops <https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.regionprops>`_
        for a list of additiona properties you can compute
        
        
    Returns
    -------
    dict
        dictionary of mean and standard deviations of 
        omi parameters for each region.

    """
    # 1. nadh_intensity
    # 2. nadh_a1
    # 3. nadh_a2
    # 4. nadh_t1
    # 5. nadh_t2
    # 6. nadh_tau_mean
    # 7. fad_intensity
    # 8. fad_a1
    # 9. fad_a2
    # 10. fad_t1
    # 11. fad_t2
    # 12. fad_tau_mean
    # 13. redox_ratio

    ##%%
    # compute composite images not generated by SPCImage, also convert a1/a2 to percent
    if (
        im_nadh_a1 is not None
        and im_nadh_a2 is not None
        and im_nadh_t1 is not None
        and im_nadh_t2 is not None
    ):
        im_nadh_tau_mean = (im_nadh_a1 / 100 * im_nadh_t1) + (
            im_nadh_a2 / 100 * im_nadh_t2
        )
    else:
        im_nadh_tau_mean = None

    if (
        im_fad_a1 is not None
        and im_fad_a2 is not None
        and im_fad_t1 is not None
        and im_fad_t2 is not None
    ):
        im_fad_tau_mean = (im_fad_a1 / 100 * im_fad_t1) + (im_fad_a2 / 100 * im_fad_t2)
    else:
        im_fad_tau_mean = None

    # COMPUTE REDOX RATIO
    # create masked arrays so it doesn't affect averaging
    if im_fad_intensity is not None and im_nadh_intensity is not None:
        labels_inverted = np.invert(label_image.astype(bool))
        masked_im_fad_intensity = ma.masked_array(
            im_fad_intensity, mask=labels_inverted
        )
        masked_im_nadh_intensity = ma.masked_array(
            im_nadh_intensity, mask=labels_inverted
        )

        # compute both types of redox ratios
        im_redox_ratio = masked_im_nadh_intensity / masked_im_fad_intensity
        im_redox_ratio_norm = masked_im_nadh_intensity / (
            masked_im_fad_intensity + masked_im_nadh_intensity
        )
    else:
        im_redox_ratio = None
        im_redox_ratio_norm = None

    # COMPUTE FLIRR
    # fluorescence lifetime imaging redox ratio aka FLIRR
    # labels mask image should not include any im_fad_a1 zeros
    # apply mask then compute to avoid Infinite values

    if im_fad_a1 is not None and im_nadh_a2 is not None:
        masked_im_fad_a1 = ma.masked_array(im_fad_a1, mask=labels_inverted)
        masked_im_nadh_a2 = ma.masked_array(im_nadh_a2, mask=labels_inverted)
        im_flirr = (masked_im_nadh_a2 / 100) / (
            masked_im_fad_a1 / 100
        )  # bound portions of NADH/FAD

        if im_flirr.any() == np.inf:
            print(
                "regionprops_omi: INF values found in FLIRR image, they will be set to zero"
            )
            im_flirr[im_flirr == np.inf] = 0
    else:
        im_flirr = None

    # define extra functions for properties
    def stdev(roi, intensity):
        inverted_roi = np.invert(roi.astype(bool))
        masked_image = ma.masked_array(intensity, mask=inverted_roi)
        return np.std(masked_image)

    def chi_median(roi, intensity):
        inverted_roi = np.invert(roi.astype(bool))
        masked_image = ma.masked_array(intensity, mask=inverted_roi)
        # plt.imshow(masked_image, vmax=1.5)
        return ma.median(masked_image)  # np.median looks at masked values

    extra_properties = [stdev, chi_median]

    # COMPUTE REGIONPROPS JUST ON BINARY IMAGE
    mask_props = regionprops(label_image)

    dict_regionprops = {}

    # COMPUTE EQUALLY WEIGHTED PARAMETERS AND ASSEMBLE DICT OF REGIONPROPS
    if im_nadh_intensity is not None:
        nadh_intensity = regionprops(
            label_image, im_nadh_intensity, extra_properties=extra_properties
        )
        dict_regionprops["nadh_intensity"] = nadh_intensity

    if im_nadh_a1 is not None:
        nadh_a1 = regionprops(
            label_image, im_nadh_a1, extra_properties=extra_properties
        )
        dict_regionprops["nadh_a1"] = nadh_a1

    if im_nadh_a2 is not None:
        nadh_a2 = regionprops(
            label_image, im_nadh_a2, extra_properties=extra_properties
        )
        dict_regionprops["nadh_a2"] = nadh_a2

    if im_nadh_t1 is not None:
        nadh_t1 = regionprops(
            label_image, im_nadh_t1, extra_properties=extra_properties
        )
        dict_regionprops["nadh_t1"] = nadh_t1

    if im_nadh_t2 is not None:
        nadh_t2 = regionprops(
            label_image, im_nadh_t2, extra_properties=extra_properties
        )
        dict_regionprops["nadh_t2"] = nadh_t2

    if im_nadh_tau_mean is not None:
        nadh_tau_mean = regionprops(
            label_image, im_nadh_tau_mean, extra_properties=extra_properties
        )
        dict_regionprops["nadh_tau_mean"] = nadh_tau_mean

    if im_fad_intensity is not None:
        fad_intensity = regionprops(
            label_image, im_fad_intensity, extra_properties=extra_properties
        )
        dict_regionprops["fad_intensity"] = fad_intensity

    if im_fad_a1 is not None:
        fad_a1 = regionprops(label_image, im_fad_a1, extra_properties=extra_properties)
        dict_regionprops["fad_a1"] = fad_a1

    if im_fad_a2 is not None:
        fad_a2 = regionprops(label_image, im_fad_a2, extra_properties=extra_properties)
        dict_regionprops["fad_a2"] = fad_a2

    if im_fad_t1 is not None:
        fad_t1 = regionprops(label_image, im_fad_t1, extra_properties=extra_properties)
        dict_regionprops["fad_t1"] = fad_t1

    if im_fad_t2 is not None:
        fad_t2 = regionprops(label_image, im_fad_t2, extra_properties=extra_properties)
        dict_regionprops["fad_t2"] = fad_t2

    if im_fad_tau_mean is not None:
        fad_tau_mean = regionprops(
            label_image, im_fad_tau_mean, extra_properties=extra_properties
        )
        dict_regionprops["fad_tau_mean"] = fad_tau_mean

    if im_redox_ratio is not None:
        redox_ratio = regionprops(
            label_image, im_redox_ratio, extra_properties=extra_properties
        )
        dict_regionprops["redox_ratio"] = redox_ratio

    if im_redox_ratio_norm is not None:
        redox_ratio_norm = regionprops(
            label_image, im_redox_ratio_norm, extra_properties=extra_properties
        )
        dict_regionprops["redox_ratio_norm"] = redox_ratio_norm

    if im_flirr is not None:
        flirr = regionprops(label_image, im_flirr, extra_properties=extra_properties)
        dict_regionprops["flirr"] = flirr

    # add chi squared values if passed in
    bool_has_chi_images = False
    if im_nadh_chi is not None and im_fad_chi is not None:
        bool_has_chi_images = True
        nadh_chi = regionprops(
            label_image, im_nadh_chi, extra_properties=extra_properties
        )
        fad_chi = regionprops(
            label_image, im_fad_chi, extra_properties=extra_properties
        )
        dict_regionprops["nadh_chi"] = nadh_chi  # add regionprops
        dict_regionprops["fad_chi"] = fad_chi  # add regionprops

    # assemble dictionary of omi parameters
    dict_omi = {}
    for rp_key in dict_regionprops.keys():  # iterate through each images regionprops
        pass
        for region in dict_regionprops[rp_key]:  # iterate through region in regionprops
            pass
            # print(region)
            dict_key_name = f"{image_id}_{region.label}"  # generate unique key for region in this image
            if not dict_key_name in dict_omi.keys():  # add region dict if needed
                pass
                dict_omi[dict_key_name] = {}  # add new dict for this label
                dict_omi[dict_key_name]["mask_label"] = int(
                    region.label
                )  # save label value

            # save equally weighted parameters
            dict_omi[dict_key_name][f"{rp_key}_mean"] = region.mean_intensity
            dict_omi[dict_key_name][f"{rp_key}_stdev"] = region.stdev

            # save chi squared median value
            if bool_has_chi_images and rp_key == "nadh_chi" or rp_key == "fad_chi":
                dict_omi[dict_key_name][f"{rp_key}_median"] = region.chi_median

            ### COMPUTE INTENSITY WEIGHTED VALUES

            # make a list of files it makes sense to compute weighted values for
            list_valid_intensity_weights = [
                "nadh_a1",
                "nadh_a2",
                "nadh_t1",
                "nadh_t2",
                "nadh_tau_mean",
                "fad_a1",
                "fad_a2",
                "fad_t1",
                "fad_t2",
                "fad_tau_mean",
            ]

            if (
                rp_key in list_valid_intensity_weights
            ):  # iterate through valid lifetime images
                pass
                # select proper intensity image to weigh by, find roi by region.label
                if "fad" in rp_key and im_fad_intensity is not None:
                    im_intensity_region = [
                        r for r in fad_intensity if r.label == region.label
                    ][
                        0
                    ]  # should be one region

                elif "nadh" in rp_key and im_nadh_intensity is not None:
                    im_intensity_region = [
                        r for r in nadh_intensity if r.label == region.label
                    ][
                        0
                    ]  # should be one region

                else:
                    continue  # skip this set
                # gather other things needed for intensity weighted
                binary = region.image
                inverted_binary = np.invert(binary)

                im_lifetime_masked = ma.masked_array(
                    region.intensity_image, mask=inverted_binary
                )
                im_intensity_masked = ma.masked_array(
                    im_intensity_region.intensity_image, mask=inverted_binary
                )

                # Kayvans way of avg weighing image
                intensity_weighted_mean = np.sum(
                    im_lifetime_masked * im_intensity_masked
                ) / np.sum(im_intensity_masked)

                # https://stackoverflow.com/questions/2413522/weighted-standard-deviation-in-numpy
                def weighted_avg_and_std(values, weights):
                    """
                    Return the weighted average and standard deviation.
                    values, weights -- Numpy ndarrays with the same shape.
                    """
                    average = np.average(values, weights=weights)

                    # Fast and numerically precise:
                    variance = np.average((values - average) ** 2, weights=weights)
                    return (average, math.sqrt(variance))

                weighted_mean, weighted_stdev = weighted_avg_and_std(
                    im_lifetime_masked, im_intensity_masked
                )
                assert weighted_mean == intensity_weighted_mean

                # save values
                dict_omi[dict_key_name][
                    f"{rp_key}_intensity_weighted_mean"
                ] = intensity_weighted_mean
                dict_omi[dict_key_name][
                    f"{rp_key}_intensity_weighted_stdev"
                ] = weighted_stdev

            ### Done adding intensity weighted params

    # ADD OTHER MASK REGIONPROP VALUES
    if other_props is not None and len(other_props) != 0:
        for region in mask_props:  # iterate through region in regionprops
            pass
            dict_key_name = f"{image_id}_{region.label}"  # generate unique key for region in this image
            # mask_region = [r for r in mask_props if r.label == region.label][0]
            for prop in other_props:
                dict_omi[dict_key_name][prop] = region[prop]

    #%%
    # dictionary of omi features could be df if we wanted to
    return dict_omi


#%%

if __name__ == "__main__":

    from pathlib import Path

    import pandas as pd

    from cell_analysis_tools.io import load_image

    path_dictionaries = Path(
        r"Z:\0-Projects and Experiments\RD - redox_ratio_development\Data Combined + QC Complete\0-dictionaries"
    )

    list_csv_files = list(path_dictionaries.glob("*"))

    csv_dict = pd.read_csv(list_csv_files[0])
    csv_dict.index.name = "base_name"

    test_dict = csv_dict.iloc[0]

    #%%

    omi_props = regionprops_omi(
        image_id="test_image",
        label_image=load_image(Path(test_dict.mask_cell)),
        im_nadh_intensity=load_image(Path(test_dict.nadh_photons)),
        im_nadh_a1=load_image(Path(test_dict.nadh_a1)),
        im_nadh_a2=load_image(Path(test_dict.nadh_a2)),
        im_nadh_t1=load_image(Path(test_dict.nadh_t1)),
        im_nadh_t2=load_image(Path(test_dict.nadh_t2)),
        im_fad_intensity=load_image(Path(test_dict.fad_photons)),
        im_fad_a1=load_image(Path(test_dict.fad_a1)),
        im_fad_a2=load_image(Path(test_dict.fad_a2)),
        im_fad_t1=load_image(Path(test_dict.fad_t1)),
        im_fad_t2=load_image(Path(test_dict.fad_t2)),
    )

    # image_id="test_image"
    # label_image = load_image(Path(test_dict.mask_cell))
    # im_nadh_intensity = load_image(Path(test_dict.nadh_photons))
    # im_nadh_a1 = load_image(Path(test_dict.nadh_a1))
    # im_nadh_a2 = None #load_image(Path(test_dict.nadh_a2))
    # im_nadh_t1 = None #load_image(Path(test_dict.nadh_t1))
    # im_nadh_t2 = None #load_image(Path(test_dict.nadh_t2))
    # im_fad_intensity =None #load_image(Path(test_dict.fad_photons))
    # im_fad_a1 = load_image(Path(test_dict.fad_a1))
    # im_fad_a2 = load_image(Path(test_dict.fad_a2))
    # im_fad_t1 = load_image(Path(test_dict.fad_t1))
    # im_fad_t2 = load_image(Path(test_dict.fad_t2))
    # im_nadh_chi = None
    # other_props = None

    pprint(omi_props)
    df = pd.DataFrame(omi_props).transpose()



from .fractal_dimension.fractal_dim_binary import fractal_dimension_binary
from .fractal_dimension.fractal_dim_gray import fractal_dimension_gray
from .regionprops import regionprops, regionprops_table

__all__ = [
    "regionprops",
    "regionprops_table",
    "fractal_dimension_gray",
    "fractal_dimension_binary",
]
